"""
Create aggregate of Lifting and Cost instances to be passed on to the main executable.
"""

import os as os
from dataclasses import dataclass, field

from pyscnomics.io.spreadsheet import Spreadsheet
from pyscnomics.econ.selection import FluidType
from pyscnomics.econ.revenue import Lifting
from pyscnomics.econ.costs import CapitalCost, Intangible, OPEX, ASR


class SpreadsheetException(Exception):
    """ Exception to raise for a misuse of Spreadsheet class """

    pass


class AggregateException(Exception):
    """ Exception to raise for a misuse of Aggregate class """

    pass


@dataclass
class Aggregate(Spreadsheet):
    """
    Prepares an aggregate of lifting data (oil, condensate, gas, LPG propane, LPG butane,
    sulfur, electricity, and CO2) and cost data (tangible, intangible, opex, ASR) to be
    passed on to the subsequent economic analysis.
    """

    # Attributes associated with aggregates of lifting data
    oil_lifting_aggregate: dict | tuple[Lifting] = field(default=None, init=False)
    condensate_lifting_aggregate: dict | tuple[Lifting] = field(default=None, init=False)
    gas_lifting_aggregate: dict | tuple[Lifting] = field(default=None, init=False)
    lpg_propane_lifting_aggregate: dict | tuple[Lifting] = field(default=None, init=False)
    lpg_butane_lifting_aggregate: dict | tuple[Lifting] = field(default=None, init=False)
    sulfur_lifting_aggregate: dict | tuple[Lifting] = field(default=None, init=False)
    electricity_lifting_aggregate: dict | tuple[Lifting] = field(default=None, init=False)
    co2_lifting_aggregate: dict | tuple[Lifting] = field(default=None, init=False)

    # Attributes associated with aggregates of cost data
    capital_cost_aggregate: dict | tuple[CapitalCost] = field(default=None, init=False)
    intangible_cost_aggregate: dict | tuple[Intangible] = field(default=None, init=False)
    opex_aggregate: dict | tuple[OPEX] = field(default=None, init=False)
    asr_cost_aggregate: dict | tuple[ASR] = field(default=None, init=False)

    # Attributes associated with aggregate of total lifting data
    oil_lifting_aggregate_total: dict | tuple[Lifting] = field(default=None, init=False)
    gas_lifting_aggregate_total: dict | tuple[Lifting] = field(default=None, init=False)

    # Attributes associated with PSC transition
    psc_regimes: list = field(default=None, init=False, repr=False)

    def __post_init__(self):
        # Configure attribute workbook_to_read
        if self.workbook_to_read is None:
            self.workbook_to_read = "Workbook.xlsb"

        if self.workbook_to_read is not None:
            if not isinstance(self.workbook_to_read, str):
                raise SpreadsheetException(
                    f"Excel filename must be provided in str format. "
                    f"{self.workbook_to_read} ({self.workbook_to_read.__class__.__qualname__}) "
                    f"is not a str format."
                )

        # Configure attribute directory_location
        if self.directory_location is None:
            self.load_dir = os.path.join(os.getcwd(), self.workbook_to_read)

        if self.directory_location is not None:
            self.load_dir = os.path.join(self.directory_location, self.workbook_to_read)

        # Arrange the loaded data
        self.prepare_data()

        # Prepare attributes associated with PSC transition
        self.psc_regimes = ["PSC 1", "PSC 2"]

    def _get_oil_lifting_aggregate(self) -> dict | tuple[Lifting]:
        """
        Retrieves the oil lifting aggregate based on the Production Sharing Contract (PSC) type.

        Returns
        -------
        Dict[str, Tuple[Lifting, ...]], Tuple[Lifting, ...]:
            -   If the type_of_contract is "Transition", returns a dictionary with
                PSC regimes as keys and a tuple of Lifting instances as values.
            -   If the type_of_contract is a single PSC (CR or GS), returns a tuple
                of Lifting instances.

        Notes
        -----
        (1) For single PSC case, the aggregate of oil lifting data is generated using
            a tuple comprehension of all available oil lifting data stored in
            parameter 'self.oil_lifting_data',
        (2) For PSC transition case, the aggregate of oil lifting data is stored in
            a dictionary with keys: ['PSC 1', 'PSC 2']. The value of each keys is a
            tuple of all available oil lifting data stored in parameter 'self.oil_lifting_data'
            for each corresponding PSC regimes.
        """
        # For PSC transition
        if "Transition" in self.oil_lifting_data.type_of_contract:
            start_year_combined = [
                self.general_config_data.start_date_project.year,
                self.general_config_data.start_date_project_second.year,
            ]

            end_year_combined = [
                self.general_config_data.end_date_project.year,
                self.general_config_data.end_date_project_second.year,
            ]

            oil_lifting_aggr = {
                psc: tuple(
                    [
                        Lifting(
                            start_year=start_year_combined[i],
                            end_year=end_year_combined[i],
                            lifting_rate=self.oil_lifting_data.oil_lifting_rate[ws][psc],
                            price=self.oil_lifting_data.oil_price[ws][psc],
                            prod_year=self.oil_lifting_data.prod_year[ws][psc],
                            fluid_type=FluidType.OIL,
                        ) for ws in self.oil_lifting_data.prod_year.keys()
                    ]
                )
                for i, psc in enumerate(self.psc_regimes)
            }

        # For single PSC (CR or GS)
        else:
            oil_lifting_aggr = tuple(
                [
                    Lifting(
                        start_year=self.general_config_data.start_date_project.year,
                        end_year=self.general_config_data.end_date_project.year,
                        lifting_rate=self.oil_lifting_data.oil_lifting_rate[ws],
                        price=self.oil_lifting_data.oil_price[ws],
                        prod_year=self.oil_lifting_data.prod_year[ws],
                        fluid_type=FluidType.OIL,
                    ) for ws in self.oil_lifting_data.prod_year.keys()
                ]
            )

        return oil_lifting_aggr

    def _get_condensate_lifting_aggregate(self) -> dict | tuple[Lifting]:
        """
        Retrieves the condensate lifting aggregate based on the Production
        Sharing Contract (PSC) type.

        Returns
        -------
        Dict[str, Tuple[Lifting, ...]], Tuple[Lifting, ...]:
            -   If the type_of_contract is 'Transition', returns a dictionary with
                PSC regimes as keys and a tuple of Lifting instances as values.
            -   If the type_of_contract is a single PSC (CR or GS), returns a tuple
                of Lifting instances.

        Notes
        -----
        (1) For single PSC case, the aggregate of condensate lifting data is generated using
            a tuple comprehension of all available condensate lifting data stored in
            parameter 'self.oil_lifting_data',
        (2) For PSC transition case, the aggregate of condensate lifting data is stored in
            a dictionary with keys: ['PSC 1', 'PSC 2']. The value of each keys is a
            tuple of all available condensate lifting data stored in parameter 'self.oil_lifting_data'
            for each corresponding PSC regimes.
        """
        # For PSC transition
        if "Transition" in self.oil_lifting_data.type_of_contract:
            start_year_combined = [
                self.general_config_data.start_date_project.year,
                self.general_config_data.start_date_project_second.year,
            ]

            end_year_combined = [
                self.general_config_data.end_date_project.year,
                self.general_config_data.end_date_project_second.year,
            ]

            condensate_lifting_aggr = {
                psc: tuple(
                    [
                        Lifting(
                            start_year=start_year_combined[i],
                            end_year=end_year_combined[i],
                            lifting_rate=self.oil_lifting_data.condensate_lifting_rate[ws][psc],
                            price=self.oil_lifting_data.condensate_price[ws][psc],
                            prod_year=self.oil_lifting_data.prod_year[ws][psc],
                            fluid_type=FluidType.OIL,
                        )
                        for ws in self.oil_lifting_data.prod_year.keys()
                    ]
                )
                for i, psc in enumerate(self.psc_regimes)
            }

        # For single PSC (CR or GS)
        else:
            condensate_lifting_aggr = tuple(
                [
                    Lifting(
                        start_year=self.general_config_data.start_date_project.year,
                        end_year=self.general_config_data.end_date_project.year,
                        lifting_rate=self.oil_lifting_data.condensate_lifting_rate[ws],
                        price=self.oil_lifting_data.condensate_price[ws],
                        prod_year=self.oil_lifting_data.prod_year[ws],
                        fluid_type=FluidType.OIL,
                    ) for ws in self.oil_lifting_data.prod_year.keys()
                ]
            )

        return condensate_lifting_aggr

    def _get_gas_lifting_aggregate(self) -> dict | tuple[Lifting]:
        """
        Retrieves the gas lifting aggregate based on the Production Sharing Contract (PSC) type.

        Returns
        -------
        Dict[str, Tuple[Lifting, ...]], Tuple[Lifting, ...]:
            -   If the type_of_contract is "Transition," returns a dictionary with
                PSC regimes as keys and a tuple of Lifting instances as values.
            -   If the type_of_contract is a single PSC (CR or GS), returns a tuple
                of Lifting instances.

        Notes
        -----
        (1) For single PSC case, the aggregate of gas lifting data is generated using
            a tuple comprehension of all available gas lifting data stored in
            parameter 'self.gas_lifting_data',
        (2) For PSC transition case, the aggregate of gas lifting data is stored in
            a dictionary with keys: ['PSC 1', 'PSC 2']. The value of each keys is a
            tuple of all available gas lifting data stored in parameter 'self.gas_lifting_data'
            for each corresponding PSC regimes.
        """
        gsa_number = self.general_config_data.gsa_number
        gsa_variables = ["GSA {0}".format(i + 1) for i in range(gsa_number)]

        # For PSC transition
        if "Transition" in self.gas_lifting_data.type_of_contract:
            start_year_combined = [
                self.general_config_data.start_date_project.year,
                self.general_config_data.start_date_project_second.year,
            ]

            end_year_combined = [
                self.general_config_data.end_date_project.year,
                self.general_config_data.end_date_project_second.year,
            ]

            gas_lifting_aggr = {
                psc: tuple(
                    [
                        Lifting(
                            start_year=start_year_combined[i],
                            end_year=end_year_combined[i],
                            lifting_rate=self.gas_lifting_data.lifting_rate[ws][gsa][psc],
                            price=self.gas_lifting_data.price[ws][gsa][psc],
                            prod_year=self.gas_lifting_data.prod_year[ws][psc],
                            fluid_type=FluidType.GAS,
                            ghv=self.gas_lifting_data.ghv[ws][gsa][psc],
                            prod_rate=self.gas_lifting_data.prod_rate[ws][psc],
                        )
                        for ws in self.gas_lifting_data.prod_year.keys()
                        for gsa in gsa_variables
                    ]
                )
                for i, psc in enumerate(self.psc_regimes)
            }

        # For single PSC (CR or GS)
        else:
            gas_lifting_aggr = tuple(
                [
                    Lifting(
                        start_year=self.general_config_data.start_date_project.year,
                        end_year=self.general_config_data.end_date_project.year,
                        lifting_rate=self.gas_lifting_data.lifting_rate[ws][gsa],
                        price=self.gas_lifting_data.price[ws][gsa],
                        prod_year=self.gas_lifting_data.prod_year[ws],
                        fluid_type=FluidType.GAS,
                        ghv=self.gas_lifting_data.ghv[ws][gsa],
                        prod_rate=self.gas_lifting_data.prod_rate[ws],
                    )
                    for ws in self.gas_lifting_data.prod_year.keys()
                    for gsa in gsa_variables
                ]
            )

        return gas_lifting_aggr

    def _get_lpg_propane_lifting_aggregate(self) -> dict | tuple[Lifting]:
        """
        Retrieves the LPG propane lifting aggregate based on the Production
        Sharing Contract (PSC) type.

        Returns
        -------
        Dict[str, Tuple[Lifting, ...]], Tuple[Lifting, ...]:
            -   If the type_of_contract is 'Transition', returns a dictionary with
                PSC regimes as keys and a tuple of Lifting instances as values.
            -   If the type_of_contract is a single PSC (CR or GS), returns a tuple
                of Lifting instances.

        Notes
        -----
        (1) For a single PSC case, the aggregate of LPG propane lifting data is generated using
            a tuple comprehension of all available LPG propane lifting data stored in
            parameter 'self.lpg_propane_lifting_data'.
        (2) For a PSC transition case, the aggregate of LPG propane lifting data is stored in
            a dictionary with keys: ['PSC 1', 'PSC 2']. The value of each key is a
            tuple of all available LPG propane lifting data stored in parameter
            'self.lpg_propane_lifting_data' for each corresponding PSC regime.
        """
        # For PSC transition
        if "Transition" in self.lpg_propane_lifting_data.type_of_contract:
            start_year_combined = [
                self.general_config_data.start_date_project.year,
                self.general_config_data.start_date_project_second.year,
            ]

            end_year_combined = [
                self.general_config_data.end_date_project.year,
                self.general_config_data.end_date_project_second.year,
            ]

            lpg_propane_lifting_aggr = {
                psc: tuple(
                    [
                        Lifting(
                            start_year=start_year_combined[i],
                            end_year=end_year_combined[i],
                            lifting_rate=self.lpg_propane_lifting_data.lifting_rate[ws][psc],
                            price=self.lpg_propane_lifting_data.price[ws][psc],
                            prod_year=self.lpg_propane_lifting_data.prod_year[ws][psc],
                            fluid_type=FluidType.GAS,
                        )
                        for ws in self.lpg_propane_lifting_data.prod_year.keys()
                    ]
                )
                for i, psc in enumerate(self.psc_regimes)
            }

        # For single PSC (CR or GS)
        else:
            lpg_propane_lifting_aggr = tuple(
                [
                    Lifting(
                        start_year=self.general_config_data.start_date_project.year,
                        end_year=self.general_config_data.end_date_project.year,
                        lifting_rate=self.lpg_propane_lifting_data.lifting_rate[ws],
                        price=self.lpg_propane_lifting_data.price[ws],
                        prod_year=self.lpg_propane_lifting_data.prod_year[ws],
                        fluid_type=FluidType.GAS,
                    )
                    for ws in self.lpg_propane_lifting_data.prod_year.keys()
                ]
            )

        return lpg_propane_lifting_aggr

    def _get_lpg_butane_lifting_aggregate(self) -> dict | tuple[Lifting]:
        """
        Retrieves the LPG butane lifting aggregate based on the Production
        Sharing Contract (PSC) type.

        Returns
        -------
        Dict[str, Tuple[Lifting, ...]], Tuple[Lifting, ...]:
            -   If the type_of_contract is 'Transition', returns a dictionary with
                PSC regimes as keys and a tuple of Lifting instances as values.
            -   If the type_of_contract is a single PSC (CR or GS), returns a tuple
                of Lifting instances.

        Notes
        -----
        (1) For a single PSC case, the aggregate of LPG butane lifting data is generated using
            a tuple comprehension of all available LPG butane lifting data stored in
            parameter 'self.lpg_butane_lifting_data'.
        (2) For a PSC transition case, the aggregate of LPG butane lifting data is stored in
            a dictionary with keys: ['PSC 1', 'PSC 2']. The value of each key is a
            tuple of all available LPG butane lifting data stored in parameter
            'self.lpg_butane_lifting_data' for each corresponding PSC regime.
        """
        # For PSC transition
        if "Transition" in self.lpg_butane_lifting_data.type_of_contract:
            start_year_combined = [
                self.general_config_data.start_date_project.year,
                self.general_config_data.start_date_project_second.year,
            ]

            end_year_combined = [
                self.general_config_data.end_date_project.year,
                self.general_config_data.end_date_project_second.year,
            ]

            lpg_butane_lifting_aggr = {
                psc: tuple(
                    [
                        Lifting(
                            start_year=start_year_combined[i],
                            end_year=end_year_combined[i],
                            lifting_rate=self.lpg_butane_lifting_data.lifting_rate[ws][psc],
                            price=self.lpg_butane_lifting_data.price[ws][psc],
                            prod_year=self.lpg_butane_lifting_data.prod_year[ws][psc],
                            fluid_type=FluidType.GAS,
                        )
                        for ws in self.lpg_butane_lifting_data.prod_year.keys()
                    ]
                )
                for i, psc in enumerate(self.psc_regimes)
            }

        # For single PSC (CR or GS)
        else:
            lpg_butane_lifting_aggr = tuple(
                [
                    Lifting(
                        start_year=self.general_config_data.start_date_project.year,
                        end_year=self.general_config_data.end_date_project.year,
                        lifting_rate=self.lpg_butane_lifting_data.lifting_rate[ws],
                        price=self.lpg_butane_lifting_data.price[ws],
                        prod_year=self.lpg_butane_lifting_data.prod_year[ws],
                        fluid_type=FluidType.GAS,
                    )
                    for ws in self.lpg_butane_lifting_data.prod_year.keys()
                ]
            )

        return lpg_butane_lifting_aggr

    def _get_sulfur_lifting_aggregate(self) -> dict | tuple[Lifting]:
        """
        Retrieves the sulfur lifting aggregate based on the Production
        Sharing Contract (PSC) type.

        Returns
        -------
        Dict[str, Tuple[Lifting, ...]], Tuple[Lifting, ...]:
            -   If the type_of_contract is 'Transition', returns a dictionary with
                PSC regimes as keys and a tuple of Lifting instances as values.
            -   If the type_of_contract is a single PSC (CR or GS), returns a tuple
                of Lifting instances.

        Notes
        -----
        (1) For a single PSC case, the aggregate of sulfur lifting data is generated using
            a tuple comprehension of all available sulfur lifting data stored in
            parameter 'self.sulfur_lifting_data'.
        (2) For a PSC transition case, the aggregate of sulfur lifting data is stored in
            a dictionary with keys: ['PSC 1', 'PSC 2']. The value of each key is a
            tuple of all available sulfur lifting data stored in parameter
            'self.sulfur_lifting_data' for each corresponding PSC regime.
        """
        # For PSC transition
        if "Transition" in self.sulfur_lifting_data.type_of_contract:
            start_year_combined = [
                self.general_config_data.start_date_project.year,
                self.general_config_data.start_date_project_second.year,
            ]

            end_year_combined = [
                self.general_config_data.end_date_project.year,
                self.general_config_data.end_date_project_second.year,
            ]

            sulfur_lifting_aggr = {
                psc: tuple(
                    [
                        Lifting(
                            start_year=start_year_combined[i],
                            end_year=end_year_combined[i],
                            lifting_rate=self.sulfur_lifting_data.lifting_rate[ws][psc],
                            price=self.sulfur_lifting_data.price[ws][psc],
                            prod_year=self.sulfur_lifting_data.prod_year[ws][psc],
                            fluid_type=FluidType.SULFUR,
                        )
                        for ws in self.sulfur_lifting_data.prod_year.keys()
                    ]
                )
                for i, psc in enumerate(self.psc_regimes)
            }

        # For single PSC (CR or GS)
        else:
            sulfur_lifting_aggr = tuple(
                [
                    Lifting(
                        start_year=self.general_config_data.start_date_project.year,
                        end_year=self.general_config_data.end_date_project.year,
                        lifting_rate=self.sulfur_lifting_data.lifting_rate[ws],
                        price=self.sulfur_lifting_data.price[ws],
                        prod_year=self.sulfur_lifting_data.prod_year[ws],
                        fluid_type=FluidType.SULFUR,
                    )
                    for ws in self.sulfur_lifting_data.prod_year.keys()
                ]
            )

        return sulfur_lifting_aggr

    def _get_electricity_lifting_aggregate(self) -> dict | tuple[Lifting]:
        """
        Retrieves the electricity lifting aggregate based on the Production
        Sharing Contract (PSC) type.

        Returns
        -------
        Dict[str, Tuple[Lifting, ...]], Tuple[Lifting, ...]:
            -   If the type_of_contract is 'Transition', returns a dictionary with
                PSC regimes as keys and a tuple of Lifting instances as values.
            -   If the type_of_contract is a single PSC (CR or GS), returns a tuple
                of Lifting instances.

        Notes
        -----
        (1) For a single PSC case, the aggregate of electricity lifting data is generated using
            a tuple comprehension of all available electricity lifting data stored in
            parameter 'self.electricity_lifting_data'.
        (2) For a PSC transition case, the aggregate of electricity lifting data is stored in
            a dictionary with keys: ['PSC 1', 'PSC 2']. The value of each key is a
            tuple of all available electricity lifting data stored in parameter
            'self.electricity_lifting_data' for each corresponding PSC regime.
        """
        # For PSC transition
        if "Transition" in self.electricity_lifting_data.type_of_contract:
            start_year_combined = [
                self.general_config_data.start_date_project.year,
                self.general_config_data.start_date_project_second.year,
            ]

            end_year_combined = [
                self.general_config_data.end_date_project.year,
                self.general_config_data.end_date_project_second.year,
            ]

            electricity_lifting_aggr = {
                psc: tuple(
                    [
                        Lifting(
                            start_year=start_year_combined[i],
                            end_year=end_year_combined[i],
                            lifting_rate=self.electricity_lifting_data.lifting_rate[ws][psc],
                            price=self.electricity_lifting_data.price[ws][psc],
                            prod_year=self.electricity_lifting_data.prod_year[ws][psc],
                            fluid_type=FluidType.ELECTRICITY,
                        )
                        for ws in self.electricity_lifting_data.prod_year.keys()
                    ]
                )
                for i, psc in enumerate(self.psc_regimes)
            }

        # For single PSC (CR or GS)
        else:
            electricity_lifting_aggr = tuple(
                [
                    Lifting(
                        start_year=self.general_config_data.start_date_project.year,
                        end_year=self.general_config_data.end_date_project.year,
                        lifting_rate=self.electricity_lifting_data.lifting_rate[ws],
                        price=self.electricity_lifting_data.price[ws],
                        prod_year=self.electricity_lifting_data.prod_year[ws],
                        fluid_type=FluidType.ELECTRICITY,
                    )
                    for ws in self.electricity_lifting_data.prod_year.keys()
                ]
            )

        return electricity_lifting_aggr

    def _get_co2_lifting_aggregate(self) -> dict | tuple[Lifting]:
        """
        Retrieves the co2 lifting aggregate based on the Production
        Sharing Contract (PSC) type.

        Returns
        -------
        Dict[str, Tuple[Lifting, ...]], Tuple[Lifting, ...]:
            -   If the type_of_contract is 'Transition', returns a dictionary with
                PSC regimes as keys and a tuple of Lifting instances as values.
            -   If the type_of_contract is a single PSC (CR or GS), returns a tuple
                of Lifting instances.

        Notes
        -----
        (1) For a single PSC case, the aggregate of co2 lifting data is generated using
            a tuple comprehension of all available co2 lifting data stored in
            parameter 'self.co2_lifting_data'.
        (2) For a PSC transition case, the aggregate of co2 lifting data is stored in
            a dictionary with keys: ['PSC 1', 'PSC 2']. The value of each key is a
            tuple of all available co2 lifting data stored in parameter
            'self.co2_lifting_data' for each corresponding PSC regime.
        """
        # For PSC transition
        if "Transition" in self.co2_lifting_data.type_of_contract:
            start_year_combined = [
                self.general_config_data.start_date_project.year,
                self.general_config_data.start_date_project_second.year,
            ]

            end_year_combined = [
                self.general_config_data.end_date_project.year,
                self.general_config_data.end_date_project_second.year,
            ]

            co2_lifting_aggr = {
                psc: tuple(
                    [
                        Lifting(
                            start_year=start_year_combined[i],
                            end_year=end_year_combined[i],
                            lifting_rate=self.co2_lifting_data.lifting_rate[ws][psc],
                            price=self.co2_lifting_data.price[ws][psc],
                            prod_year=self.co2_lifting_data.prod_year[ws][psc],
                            fluid_type=FluidType.CO2,
                        )
                        for ws in self.co2_lifting_data.prod_year.keys()
                    ]
                )
                for i, psc in enumerate(self.psc_regimes)
            }

        # For single PSC (CR or GS)
        else:
            co2_lifting_aggr = tuple(
                [
                    Lifting(
                        start_year=self.general_config_data.start_date_project.year,
                        end_year=self.general_config_data.end_date_project.year,
                        lifting_rate=self.co2_lifting_data.lifting_rate[ws],
                        price=self.co2_lifting_data.price[ws],
                        prod_year=self.co2_lifting_data.prod_year[ws],
                        fluid_type=FluidType.CO2,
                    )
                    for ws in self.co2_lifting_data.prod_year.keys()
                ]
            )

        return co2_lifting_aggr

    def _get_oil_lifting_aggregate_total(
        self,
        oil_lifting_aggregate: dict | tuple,
        condensate_lifting_aggregate: dict | tuple,
    ) -> dict | tuple[Lifting]:
        """
        Calculate the total oil lifting aggregate for a Production Sharing Contract (PSC).

        Parameters
        ----------
        oil_lifting_aggregate: dict | tuple
            Dictionary or tuple containing oil lifting data.
        condensate_lifting_aggregate: dict | tuple
            Dictionary or tuple containing condensate lifting data.

        Returns
        -------
        -   If the type_of_contract is 'Transition', returns a dictionary with
            PSC regimes as keys and a tuple of Lifting instances as values.
        -   If the type_of_contract is a single PSC (CR or GS), returns a tuple
            of Lifting instances.

        Notes
        -----
        -   For PSC transition, the function sums the oil and condensate lifting
            aggregates for each PSC regime.
        -   For a single PSC (CR or GS), the function simply concatenates the oil
            and condensate lifting aggregates.
        """
        # For PSC transition
        if "Transition" in self.oil_lifting_data.type_of_contract:
            oil_lifting_aggr_tot = {
                psc: oil_lifting_aggregate[psc] + condensate_lifting_aggregate[psc]
                for psc in self.psc_regimes
            }

        # For single PSC (CR or GS)
        else:
            oil_lifting_aggr_tot = oil_lifting_aggregate + condensate_lifting_aggregate

        return oil_lifting_aggr_tot

    def _get_gas_lifting_aggregate_total(
        self,
        gas_lifting_aggregate: dict | tuple,
        lpg_propane_lifting_aggregate: dict | tuple,
        lpg_butane_lifting_aggregate: dict | tuple,
    ) -> dict | tuple | AggregateException:
        """
        Calculate the total gas lifting aggregate for a Production Sharing Contract (PSC).

        Parameters
        ----------
        gas_lifting_aggregate: dict | tuple
            Dictionary or tuple containing gas lifting data.
        lpg_propane_lifting_aggregate: dict | tuple
            Dictionary or tuple containing LPG propane lifting data.
        lpg_butane_lifting_aggregate: dict | tuple
            Dictionary or tuple containing LPG butane lifting data.

        Returns
        -------
        dict | tuple | AggregateException
            -   If all contracts are transition (contain "Transition" in their type),
                returns a dictionary representing the total gas lifting aggregate for
                each PSC regime.
            -   If all contracts are single PSCs (CR or GS, without "Transition" in their type),
                returns a tuple containing the aggregated gas lifting data.
            -   If contracts mix both transition and non-transition types,
                returns an Exception.

        Notes
        -----
        -   For PSC transition, the function sums the gas, LPG propane, and LPG butane
            lifting aggregates for each PSC regime.
        -   For a single PSC (CR or GS), the function simply concatenates the gas,
            LPG propane, and LPG butane lifting aggregates.
        """
        contracts = [
            self.gas_lifting_data.type_of_contract,
            self.lpg_propane_lifting_data.type_of_contract,
            self.lpg_butane_lifting_data.type_of_contract,
        ]

        # For PSC transition
        if all(["Transition" in i for i in contracts]):
            return {
                psc: (
                    gas_lifting_aggregate[psc]
                    + lpg_propane_lifting_aggregate[psc]
                    + lpg_butane_lifting_aggregate[psc]
                )
                for psc in self.psc_regimes
            }

        # For single PSC (CR or GS)
        elif all(["Transition" not in i for i in contracts]):
            return (
                gas_lifting_aggregate
                + lpg_propane_lifting_aggregate
                + lpg_butane_lifting_aggregate
            )

        else:
            return AggregateException(
                f"Unequal type of PSC contracts for Gas, LPG propane, and LPG butane. "
                f"Gas: {self.gas_lifting_data.type_of_contract}, "
                f"LPG Propane: {self.lpg_propane_lifting_data.type_of_contract}, "
                f"LPG Butane: {self.lpg_butane_lifting_data.type_of_contract}."
            )

    def _get_capital_cost_aggregate(self) -> dict | tuple[CapitalCost]:
        """
        Retrieves the capital cost aggregate based on the Production
        Sharing Contract (PSC) type.

        Returns
        -------
        -   If PSC transition: dict
                A dictionary containing PSC regimes as keys and tuples of CapitalCost objects as values.
        -   If single PSC (CR or GS): tuple[CapitalCost]
                A tuple containing a single CapitalCost object.

        Notes
        -----
        For a PSC transition case, the aggregate of capital cost data is stored in
        a dictionary with keys: ['PSC 1', 'PSC 2']. The value of each key is a tuple
        of capital cost data stored in parameter 'self.capital_cost_data' for each
        corresponding PSC regime.
        """
        # For PSC transition
        if "Transition" in self.capital_cost_data.type_of_contract:
            start_year_combined = [
                self.general_config_data.start_date_project.year,
                self.general_config_data.start_date_project_second.year,
            ]

            end_year_combined = [
                self.general_config_data.end_date_project.year,
                self.general_config_data.end_date_project_second.year,
            ]

            capital_cost_aggr = {
                psc: tuple(
                    [
                        CapitalCost(
                            start_year=start_year_combined[i],
                            end_year=end_year_combined[i],
                            cost=self.capital_cost_data.cost[psc],
                            expense_year=self.capital_cost_data.expense_year[psc],
                            cost_allocation=self.capital_cost_data.cost_allocation[psc],
                            vat_portion=self.capital_cost_data.vat_portion[psc],
                            vat_discount=self.fiscal_config_data.vat_discount[psc],
                            lbt_portion=self.capital_cost_data.lbt_portion[psc],
                            lbt_discount=self.fiscal_config_data.lbt_discount[psc],
                            description=self.capital_cost_data.description[psc],
                            pis_year=self.capital_cost_data.pis_year[psc],
                            salvage_value=self.capital_cost_data.salvage_value[psc],
                            useful_life=self.capital_cost_data.useful_life[psc],
                            depreciation_factor=self.capital_cost_data.depreciation_factor[psc],
                            is_ic_applied=self.capital_cost_data.is_ic_applied[psc],
                        )
                    ]
                )
                for i, psc in enumerate(self.psc_regimes)
            }

        # For single PSC (CR or GS)
        else:
            capital_cost_aggr = tuple(
                [
                    CapitalCost(
                        start_year=self.general_config_data.start_date_project.year,
                        end_year=self.general_config_data.end_date_project.year,
                        cost=self.capital_cost_data.cost,
                        expense_year=self.capital_cost_data.expense_year,
                        cost_allocation=self.capital_cost_data.cost_allocation,
                        vat_portion=self.capital_cost_data.vat_portion,
                        vat_discount=self.fiscal_config_data.vat_discount,
                        lbt_portion=self.capital_cost_data.lbt_portion,
                        lbt_discount=self.fiscal_config_data.lbt_discount,
                        description=self.capital_cost_data.description,
                        pis_year=self.capital_cost_data.pis_year,
                        salvage_value=self.capital_cost_data.salvage_value,
                        useful_life=self.capital_cost_data.useful_life,
                        depreciation_factor=self.capital_cost_data.depreciation_factor,
                        is_ic_applied=self.capital_cost_data.is_ic_applied,
                    )
                ]
            )

        return capital_cost_aggr

    def _get_intangible_cost_aggregate(self) -> dict | tuple[Intangible]:
        """
        Retrieves the intangible cost aggregate based on the Production
        Sharing Contract (PSC) type.

        Returns
        -------
        -   If PSC transition: dict
                A dictionary containing PSC regimes as keys and tuples of Intangible objects as values.
        -   If single PSC (CR or GS): tuple[Intangible]
                A tuple containing a single Intangible object.

        Notes
        -----
        For a PSC transition case, the aggregate of intangible cost data is stored in
        a dictionary with keys: ['PSC 1', 'PSC 2']. The value of each key is a tuple
        of intangible cost data stored in parameter 'self.intangible_cost_data' for each
        corresponding PSC regime.
        """
        # For PSC transition
        if "Transition" in self.intangible_cost_data.type_of_contract:
            start_year_combined = [
                self.general_config_data.start_date_project.year,
                self.general_config_data.start_date_project_second.year,
            ]

            end_year_combined = [
                self.general_config_data.end_date_project.year,
                self.general_config_data.end_date_project_second.year,
            ]

            intangible_cost_aggr = {
                psc: tuple(
                    [
                        Intangible(
                            start_year=start_year_combined[i],
                            end_year=end_year_combined[i],
                            cost=self.intangible_cost_data.cost[psc],
                            expense_year=self.intangible_cost_data.expense_year[psc],
                            cost_allocation=self.intangible_cost_data.cost_allocation[psc],
                            description=self.intangible_cost_data.description[psc],
                            vat_portion=self.intangible_cost_data.vat_portion[psc],
                            vat_discount=self.fiscal_config_data.vat_discount[psc],
                            lbt_portion=self.intangible_cost_data.lbt_portion[psc],
                            lbt_discount=self.fiscal_config_data.lbt_discount[psc],
                        )
                    ]
                )
                for i, psc in enumerate(self.psc_regimes)
            }

        # For single PSC (CR or GS)
        else:
            intangible_cost_aggr = tuple(
                [
                    Intangible(
                        start_year=self.general_config_data.start_date_project.year,
                        end_year=self.general_config_data.end_date_project.year,
                        cost=self.intangible_cost_data.cost,
                        expense_year=self.intangible_cost_data.expense_year,
                        cost_allocation=self.intangible_cost_data.cost_allocation,
                        description=self.intangible_cost_data.description,
                        vat_portion=self.intangible_cost_data.vat_portion,
                        vat_discount=self.fiscal_config_data.vat_discount,
                        lbt_portion=self.intangible_cost_data.lbt_portion,
                        lbt_discount=self.fiscal_config_data.lbt_discount,
                    )
                ]
            )

        return intangible_cost_aggr

    def _get_opex_aggregate(self) -> dict | tuple[OPEX]:
        """
        Retrieves the OPEX aggregate based on the Production
        Sharing Contract (PSC) type.

        Returns
        -------
        -   If PSC transition: dict
                A dictionary containing PSC regimes as keys and tuples of OPEX objects as values.
        -   If single PSC (CR or GS): tuple[OPEX]
                A tuple containing a single OPEX object.

        Notes
        -----
        For a PSC transition case, the aggregate of OPEX data is stored in
        a dictionary with keys: ['PSC 1', 'PSC 2']. The value of each key is a tuple
        of OPEX data stored in parameter 'self.opex_data' for each
        corresponding PSC regime.
        """
        # For PSC transition
        if "Transition" in self.opex_data.type_of_contract:
            start_year_combined = [
                self.general_config_data.start_date_project.year,
                self.general_config_data.start_date_project_second.year,
            ]

            end_year_combined = [
                self.general_config_data.end_date_project.year,
                self.general_config_data.end_date_project_second.year,
            ]

            opex_aggr = {
                psc: tuple(
                    [
                        OPEX(
                            start_year=start_year_combined[i],
                            end_year=end_year_combined[i],
                            expense_year=self.opex_data.expense_year[psc],
                            cost_allocation=self.opex_data.cost_allocation[psc],
                            description=self.opex_data.description[psc],
                            vat_portion=self.opex_data.vat_portion[psc],
                            vat_discount=self.fiscal_config_data.vat_discount[psc],
                            lbt_portion=self.opex_data.lbt_portion[psc],
                            lbt_discount=self.fiscal_config_data.lbt_discount[psc],
                            fixed_cost=self.opex_data.fixed_cost[psc],
                            prod_rate=self.opex_data.prod_rate[psc],
                            cost_per_volume=self.opex_data.cost_per_volume[psc],
                        )
                    ]
                )
                for i, psc in enumerate(self.psc_regimes)
            }

        # For single PSC (CR or GS)
        else:
            opex_aggr = tuple(
                [
                    OPEX(
                        start_year=self.general_config_data.start_date_project.year,
                        end_year=self.general_config_data.end_date_project.year,
                        expense_year=self.opex_data.expense_year,
                        cost_allocation=self.opex_data.cost_allocation,
                        description=self.opex_data.description,
                        vat_portion=self.opex_data.vat_portion,
                        vat_discount=self.fiscal_config_data.vat_discount,
                        lbt_portion=self.opex_data.lbt_portion,
                        lbt_discount=self.fiscal_config_data.lbt_discount,
                        fixed_cost=self.opex_data.fixed_cost,
                        prod_rate=self.opex_data.prod_rate,
                        cost_per_volume=self.opex_data.cost_per_volume,
                    )
                ]
            )

        return opex_aggr

    def _get_asr_cost_aggregate(self) -> dict | tuple[ASR]:
        """
        Retrieves the ASR cost aggregate based on the Production
        Sharing Contract (PSC) type.

        Returns
        -------
        -   If PSC transition: dict
                A dictionary containing PSC regimes as keys and tuples of ASR objects as values.
        -   If single PSC (CR or GS): tuple[ASR]
                A tuple containing a single ASR object.

        Notes
        -----
        For a PSC transition case, the aggregate of ASR cost data is stored in
        a dictionary with keys: ['PSC 1', 'PSC 2']. The value of each key is a tuple
        of ASR cost data stored in parameter 'self.asr_cost_data' for each
        corresponding PSC regime.
        """
        # For PSC transition
        if "Transition" in self.asr_cost_data.type_of_contract:
            start_year_combined = [
                self.general_config_data.start_date_project.year,
                self.general_config_data.start_date_project_second.year,
            ]

            end_year_combined = [
                self.general_config_data.end_date_project.year,
                self.general_config_data.end_date_project_second.year,
            ]

            asr_cost_aggr = {
                psc: tuple(
                    [
                        ASR(
                            start_year=start_year_combined[i],
                            end_year=end_year_combined[i],
                            cost=self.asr_cost_data.cost[psc],
                            expense_year=self.asr_cost_data.expense_year[psc],
                            cost_allocation=self.asr_cost_data.cost_allocation[psc],
                            description=self.asr_cost_data.description[psc],
                            vat_portion=self.asr_cost_data.vat_portion[psc],
                            vat_discount=self.fiscal_config_data.vat_discount[psc],
                            lbt_portion=self.asr_cost_data.lbt_portion[psc],
                            lbt_discount=self.fiscal_config_data.lbt_discount[psc],
                        )
                    ]
                )
                for i, psc in enumerate(self.psc_regimes)
            }

        # For single PSC (CR or GS)
        else:
            asr_cost_aggr = tuple(
                [
                    ASR(
                        start_year=self.general_config_data.start_date_project.year,
                        end_year=self.general_config_data.end_date_project.year,
                        cost=self.asr_cost_data.cost,
                        expense_year=self.asr_cost_data.expense_year,
                        cost_allocation=self.asr_cost_data.cost_allocation,
                        description=self.asr_cost_data.description,
                        vat_portion=self.asr_cost_data.vat_portion,
                        vat_discount=self.fiscal_config_data.vat_discount,
                        lbt_portion=self.asr_cost_data.lbt_portion,
                        lbt_discount=self.fiscal_config_data.lbt_discount,
                    )
                ]
            )

        return asr_cost_aggr

    def fit(self) -> None:
        """
        Fit method for the Aggregate class.
        This method assigns aggregates associated with lifting data and costs data.

        Returns
        -------
        None
            This method modifies the instance attributes of Aggregate class.

        Notes
        -----
        This method serves as a comprehensive data preparation step, calling specific
        private methods to obtain and assign data based on their category.

        The attributes include lifting and costs aggregates for all available components.
        """
        # Aggregates associated with lifting data
        self.oil_lifting_aggregate = self._get_oil_lifting_aggregate()
        self.gas_lifting_aggregate = self._get_gas_lifting_aggregate()
        self.condensate_lifting_aggregate = self._get_condensate_lifting_aggregate()
        self.lpg_propane_lifting_aggregate = self._get_lpg_propane_lifting_aggregate()
        self.lpg_butane_lifting_aggregate = self._get_lpg_butane_lifting_aggregate()
        self.sulfur_lifting_aggregate = self._get_sulfur_lifting_aggregate()
        self.electricity_lifting_aggregate = self._get_electricity_lifting_aggregate()
        self.co2_lifting_aggregate = self._get_co2_lifting_aggregate()

        # Aggregates associated with total lifting data
        self.oil_lifting_aggregate_total = self._get_oil_lifting_aggregate_total(
            oil_lifting_aggregate=self.oil_lifting_aggregate,
            condensate_lifting_aggregate=self.condensate_lifting_aggregate,
        )

        self.gas_lifting_aggregate_total = self._get_gas_lifting_aggregate_total(
            gas_lifting_aggregate=self.gas_lifting_aggregate,
            lpg_propane_lifting_aggregate=self.lpg_propane_lifting_aggregate,
            lpg_butane_lifting_aggregate=self.lpg_butane_lifting_aggregate,
        )

        # Aggregates associated with costs data
        self.capital_cost_aggregate = self._get_capital_cost_aggregate()
        self.intangible_cost_aggregate = self._get_intangible_cost_aggregate()
        self.opex_aggregate = self._get_opex_aggregate()
        self.asr_cost_aggregate = self._get_asr_cost_aggregate()



"""
This test code will be changed in the version 1_2_0
"""
# ToDo: Rewrite the test code in the version 1.2.0

import numpy as np
from pyscnomics.dataset.sample import load_data, load_testing
from pyscnomics.econ.selection import FTPTaxRegime

psc = load_data(dataset_type='case3', contract_type='cost_recovery')

# Editing the CostRecovery attribute as the corresponding case 1
# FTP
psc.oil_ftp_is_available = True
psc.oil_ftp_is_shared = True
psc.oil_ftp_portion = 0.2
psc.gas_ftp_is_available = True
psc.gas_ftp_is_shared = True
psc.gas_ftp_portion = 0.2

# Split Pre Tax
psc.oil_ctr_pretax_share = 0.347222
psc.gas_ctr_pretax_share = 0.520833

# DMO
psc.oil_dmo_volume_portion = 0.25
psc.oil_dmo_fee_portion = 0.25
psc.oil_dmo_holiday_duration = 60
psc.gas_dmo_volume_portion = 0.25
psc.gas_dmo_fee_portion = 1
psc.gas_dmo_holiday_duration = 60

tax_rate = 0.424

psc.run(tax_rate=tax_rate, ftp_tax_regime=FTPTaxRegime.PRE_PDJP_20_2017)


def test_revenue():
    """
    Test module to test value of revenue calculated by pyscnomics.
    """

    oil_revenue = psc._oil_revenue
    oil_revenue_base = np.asarray(load_testing(dataset_type='case3', key='oil_revenue'))

    gas_revenue = psc._gas_revenue
    gas_revenue_base = np.asarray(load_testing(dataset_type='case3', key='gas_revenue'))

    np.testing.assert_allclose(oil_revenue, oil_revenue_base, atol=1e-6)
    np.testing.assert_allclose(gas_revenue, gas_revenue_base, atol=1e-6)


def test_depreciation():
    """
    Test module for testing the Depreciation.
    """

    oil_depreciation = psc._oil_depreciation
    oil_depreciation_base = load_testing(dataset_type='case3', key='oil_depreciation')
    gas_depreciation = psc._gas_depreciation
    gas_depreciation_base = load_testing(dataset_type='case3', key='gas_depreciation')

    np.testing.assert_allclose(oil_depreciation, oil_depreciation_base, atol=1e-6)
    np.testing.assert_allclose(gas_depreciation, gas_depreciation_base, atol=1e-6)


def test_non_capital():
    oil_non_capital = psc._oil_non_capital
    oil_non_capital_base = load_testing(dataset_type='case3', key='oil_non_capital')
    gas_non_capital = psc._gas_non_capital
    gas_non_capital_base = load_testing(dataset_type='case3', key='gas_non_capital')

    np.testing.assert_allclose(oil_non_capital, oil_non_capital_base, atol=1e-6)
    np.testing.assert_allclose(gas_non_capital, gas_non_capital_base, atol=1e-6)


def test_ftp():
    oil_ftp = psc._oil_ftp
    oil_ftp_base = load_testing(dataset_type='case3', key='oil_ftp')
    gas_ftp = psc._gas_ftp
    gas_ftp_base = load_testing(dataset_type='case3', key='gas_ftp')

    np.testing.assert_allclose(oil_ftp, oil_ftp_base, atol=1e-6)
    np.testing.assert_allclose(gas_ftp, gas_ftp_base, atol=1e-6)


def test_ftp_ctr():
    oil_ftp_ctr = psc._oil_ftp_ctr
    oil_ftp_ctr_base = load_testing(dataset_type='case3', key='oil_ftp_ctr')
    gas_ftp_ctr = psc._gas_ftp_ctr
    gas_ftp_ctr_base = load_testing(dataset_type='case3', key='gas_ftp_ctr')

    np.testing.assert_allclose(oil_ftp_ctr, oil_ftp_ctr_base, atol=1e-6)
    np.testing.assert_allclose(gas_ftp_ctr, gas_ftp_ctr_base, atol=1e-6)


def test_ftp_gov():
    oil_ftp_gov = psc._oil_ftp_gov
    oil_ftp_gov_base = load_testing(dataset_type='case3', key='oil_ftp_gov')
    gas_ftp_gov = psc._gas_ftp_gov
    gas_ftp_gov_base = load_testing(dataset_type='case3', key='gas_ftp_gov')

    np.testing.assert_allclose(oil_ftp_gov, oil_ftp_gov_base, atol=1e-6)
    np.testing.assert_allclose(gas_ftp_gov, gas_ftp_gov_base, atol=1e-6)


def test_ic():
    """
    Test module for testing the Investment Credit (IC).
    """

    oil_ic = psc._oil_ic
    oil_ic_unrecovered = psc._oil_ic_unrecovered
    oil_ic_paid = psc._oil_ic_paid
    gas_ic = psc._gas_ic
    gas_ic_unrecovered = psc._gas_ic_unrecovered
    gas_ic_paid = psc._gas_ic_paid

    oil_ic_base = load_testing(dataset_type='case3', key='oil_ic')
    oil_ic_unrecovered_base = load_testing(dataset_type='case3', key='oil_ic_unrec')
    oil_ic_paid_base = load_testing(dataset_type='case3', key='oil_ic_paid')
    gas_ic_base = load_testing(dataset_type='case3', key='gas_ic')
    gas_ic_unrecovered_base = load_testing(dataset_type='case3', key='gas_ic_unrec')
    gas_ic_paid_base = load_testing(dataset_type='case3', key='gas_ic_paid')

    np.testing.assert_allclose(oil_ic, oil_ic_base, atol=1e-6)
    np.testing.assert_allclose(oil_ic_unrecovered, oil_ic_unrecovered_base, atol=1e-6)
    np.testing.assert_allclose(oil_ic_paid, oil_ic_paid_base, atol=1e-6)
    np.testing.assert_allclose(gas_ic, gas_ic_base, atol=1e-6)
    np.testing.assert_allclose(gas_ic_unrecovered, gas_ic_unrecovered_base, atol=1e-6)
    np.testing.assert_allclose(gas_ic_paid, gas_ic_paid_base, atol=1e-6)


def test_unrec_before_transfer():
    """
    Test module for testing the Unrecovered Cost Before Transfer.
    """

    oil_unrec_before_transfer = psc._oil_unrecovered_before_transfer
    oil_unrec_before_transfer_base = load_testing(dataset_type='case3', key='oil_unrec_cost')
    gas_unrec_before_transfer = psc._gas_unrecovered_before_transfer
    gas_unrec_before_transfer_base = load_testing(dataset_type='case3', key='gas_unrec_cost')

    np.testing.assert_allclose(oil_unrec_before_transfer, oil_unrec_before_transfer_base, atol=1e-6)
    np.testing.assert_allclose(gas_unrec_before_transfer, gas_unrec_before_transfer_base, atol=1e-6)


def test_cost_tobe_recovered():
    """
    Test module for testing the Cost To Be Recover.
    """

    oil_cost_to_be_recovered = psc._oil_cost_to_be_recovered
    gas_cost_to_be_recovered = psc._gas_cost_to_be_recovered

    oil_cost_to_be_recovered_base = load_testing(dataset_type='case3', key='oil_cost_to_be_recovered')
    gas_cost_to_be_recovered_base = load_testing(dataset_type='case3', key='gas_cost_to_be_recovered')

    np.testing.assert_allclose(oil_cost_to_be_recovered, oil_cost_to_be_recovered_base, atol=1e-6)
    np.testing.assert_allclose(gas_cost_to_be_recovered, gas_cost_to_be_recovered_base, atol=1e-6)


def test_cost_recovery():
    # Calculated result
    oil_cost_recovery = psc._oil_cost_recovery
    gas_cost_recovery = psc._gas_cost_recovery

    # Expected result
    oil_cost_recovery_base = load_testing(dataset_type='case3', key='oil_cost_recovery')
    gas_cost_recovery_base = load_testing(dataset_type='case3', key='gas_cost_recovery')

    # Execute testing
    np.testing.assert_allclose(oil_cost_recovery, oil_cost_recovery_base, atol=1e-6)
    np.testing.assert_allclose(gas_cost_recovery, gas_cost_recovery_base, atol=1e-6)


def test_ets_before_transfer():
    # Calculated result
    oil_ets_before_transfer = psc._oil_ets_before_transfer
    gas_ets_before_transfer = psc._gas_ets_before_transfer

    # Expected result
    oil_ets_before_transfer_base = load_testing(dataset_type='case3', key='oil_ets_before_transfer')
    gas_ets_before_transfer_base = load_testing(dataset_type='case3', key='gas_ets_before_transfer')

    # Execute testing
    np.testing.assert_allclose(oil_ets_before_transfer, oil_ets_before_transfer_base, atol=1e-6)
    np.testing.assert_allclose(gas_ets_before_transfer, gas_ets_before_transfer_base, atol=1e-6)


def test_transfer():
    # Calculated result
    transfer_to_gas = psc._transfer_to_gas
    transfer_to_oil = psc._transfer_to_oil

    # Expected result
    transfer_to_gas_base = load_testing(dataset_type='case3', key='oil_transfer_to_gas')
    transfer_to_oil_base = load_testing(dataset_type='case3', key='gas_transfer_to_oil')

    # Execute testing
    np.testing.assert_allclose(transfer_to_gas, transfer_to_gas_base, atol=1e-6)
    np.testing.assert_allclose(transfer_to_oil, transfer_to_oil_base, atol=1e-6)


def test_unrecovered_after_transfer():
    # Calculated result
    oil_unrecovered_after_transfer = psc._oil_unrecovered_after_transfer
    gas_unrecovered_after_transfer = psc._gas_unrecovered_after_transfer

    # Expected result
    oil_unrecovered_after_transfer_base = load_testing(dataset_type='case3', key='oil_unrec_after_transfer')
    gas_unrecovered_after_transfer_base = load_testing(dataset_type='case3', key='gas_unrec_after_transfer')

    # Execute testing
    np.testing.assert_allclose(oil_unrecovered_after_transfer, oil_unrecovered_after_transfer_base, atol=1e-6)
    np.testing.assert_allclose(gas_unrecovered_after_transfer, gas_unrecovered_after_transfer_base, atol=1e-6)


def test_ets_after_transfer():
    # Calculated result
    oil_ets_after_transfer = psc._oil_ets_after_transfer
    gas_ets_after_transfer = psc._gas_ets_after_transfer

    # Expected result
    oil_ets_after_transfer_base = load_testing(dataset_type='case3', key='oil_ets_after_transfer')
    gas_ets_after_transfer_base = load_testing(dataset_type='case3', key='gas_ets_after_transfer')

    # Execute testing
    np.testing.assert_allclose(oil_ets_after_transfer, oil_ets_after_transfer_base, atol=1e-6)
    np.testing.assert_allclose(gas_ets_after_transfer, gas_ets_after_transfer_base, atol=1e-6)


def test_equity_share():
    # Calculated result
    oil_contractor_share = psc._oil_contractor_share
    oil_government_share = psc._oil_government_share

    gas_contractor_share = psc._gas_contractor_share
    gas_government_share = psc._gas_government_share

    # Expected result
    oil_contractor_share_base = load_testing(dataset_type='case3', key='oil_ctr_share')
    oil_government_share_base = load_testing(dataset_type='case3', key='oil_gov_share')
    gas_contractor_share_base = load_testing(dataset_type='case3', key='gas_ctr_share')
    gas_government_share_base = load_testing(dataset_type='case3', key='gas_gov_share')

    # Execute testing
    np.testing.assert_allclose(oil_contractor_share, oil_contractor_share_base, atol=1e-6)
    np.testing.assert_allclose(oil_government_share, oil_government_share_base, atol=1e-6)
    np.testing.assert_allclose(gas_contractor_share, gas_contractor_share_base, atol=1e-6)
    np.testing.assert_allclose(gas_government_share, gas_government_share_base, atol=1e-6)


def test_dmo():
    # Calculated result
    oil_dmo_volume = psc._oil_dmo_volume
    oil_dmo_fee = psc._oil_dmo_fee
    oil_ddmo = psc._oil_ddmo

    gas_dmo_volume = psc._gas_dmo_volume
    gas_dmo_fee = psc._gas_dmo_fee
    gas_ddmo = psc._gas_ddmo

    # Expected result
    oil_dmo_volume_base = load_testing(dataset_type='case3', key='oil_dmo_volume')
    oil_dmo_fee_base = load_testing(dataset_type='case3', key='oil_dmo_fee')
    oil_ddmo_base = load_testing(dataset_type='case3', key='oil_ddmo')

    gas_dmo_volume_base = load_testing(dataset_type='case3', key='gas_dmo_volume')
    gas_dmo_fee_base = load_testing(dataset_type='case3', key='gas_dmo_fee')
    gas_ddmo_base = load_testing(dataset_type='case3', key='gas_ddmo')

    # Execute testing
    np.testing.assert_allclose(oil_dmo_volume, oil_dmo_volume_base, atol=1e-6)
    np.testing.assert_allclose(oil_dmo_fee, oil_dmo_fee_base, atol=1e-6)
    np.testing.assert_allclose(oil_ddmo, oil_ddmo_base, atol=1e-6)

    np.testing.assert_allclose(gas_dmo_volume, gas_dmo_volume_base, atol=1e-6)
    np.testing.assert_allclose(gas_dmo_fee, gas_dmo_fee_base, atol=1e-6)
    np.testing.assert_allclose(gas_ddmo, gas_ddmo_base, atol=1e-6)


def test_taxable_income():
    # Calculated result
    oil_taxable_income = psc._oil_taxable_income
    gas_taxable_income = psc._gas_taxable_income

    # Expected result
    oil_taxable_income_base = load_testing(dataset_type='case3', key='oil_taxable_income')
    gas_taxable_income_base = load_testing(dataset_type='case3', key='gas_taxable_income')

    # Execute testing
    np.testing.assert_allclose(oil_taxable_income, oil_taxable_income_base, atol=1e-6)
    np.testing.assert_allclose(gas_taxable_income, gas_taxable_income_base, atol=1e-6)


def test_tax():
    # Calculated result
    oil_tax = psc._oil_tax_payment
    gas_tax = psc._gas_tax_payment

    # Expected result
    oil_tax_base = load_testing(dataset_type='case3', key='oil_tax')
    gas_tax_base = load_testing(dataset_type='case3', key='gas_tax')

    # Execute testing
    np.testing.assert_allclose(oil_tax, oil_tax_base, atol=1e-6)
    np.testing.assert_allclose(gas_tax, gas_tax_base, atol=1e-6)


def test_ctr_net_share():
    # Calculated result
    oil_ctr_share_after_tax = psc._oil_ctr_net_share
    gas_ctr_share_after_tax = psc._gas_ctr_net_share

    # Expected result
    oil_ctr_share_after_tax_base = load_testing(dataset_type='case3', key='oil_ctr_net_share')
    gas_ctr_share_after_tax_base = load_testing(dataset_type='case3', key='gas_ctr_net_share')

    # Execute testing
    np.testing.assert_allclose(oil_ctr_share_after_tax, oil_ctr_share_after_tax_base, atol=1e-6)
    np.testing.assert_allclose(gas_ctr_share_after_tax, gas_ctr_share_after_tax_base, atol=1e-6)


def test_ctr_take():
    # Calculated result
    oil_ctr_share_after_tax = psc._oil_ctr_net_share
    gas_ctr_share_after_tax = psc._gas_ctr_net_share

    # Expected result
    oil_ctr_share_after_tax_base = load_testing(dataset_type='case3', key='oil_ctr_net_share')
    gas_ctr_share_after_tax_base = load_testing(dataset_type='case3', key='gas_ctr_net_share')

    # Execute testing
    np.testing.assert_allclose(oil_ctr_share_after_tax, oil_ctr_share_after_tax_base, atol=1e-6)
    np.testing.assert_allclose(gas_ctr_share_after_tax, gas_ctr_share_after_tax_base, atol=1e-6)


def test_ctr_cashflow():
    # Calculated result
    oil_ctr_cashflow = psc._oil_cashflow
    gas_ctr_cashflow = psc._gas_cashflow

    # Expected result
    oil_ctr_cashflow_base = load_testing(dataset_type='case3', key='oil_ctr_cashflow')
    gas_ctr_cashflow_base = load_testing(dataset_type='case3', key='gas_ctr_cashflow')

    # Execute testing
    np.testing.assert_allclose(oil_ctr_cashflow, oil_ctr_cashflow_base, atol=1e-6)
    np.testing.assert_allclose(gas_ctr_cashflow, gas_ctr_cashflow_base, atol=1e-6)


def test_goi_take():
    # Calculated result
    oil_goi_take = psc._oil_government_take
    gas_goi_take = psc._gas_government_take

    # Expected result
    oil_goi_take_base = load_testing(dataset_type='case3', key='oil_goi_take')
    gas_goi_take_base = load_testing(dataset_type='case3', key='gas_goi_take')

    # Execute testing
    np.testing.assert_allclose(oil_goi_take, oil_goi_take_base, atol=1e-6)
    np.testing.assert_allclose(gas_goi_take, gas_goi_take_base, atol=1e-6)


def test_consistency():
    # Top side result
    gross_revenue_topside = psc._oil_revenue + psc._gas_revenue

    # Bottom side result
    gross_revenue_botside = psc._oil_contractor_take + psc._gas_contractor_take + psc._oil_government_take + psc._gas_government_take

    # Execute testing
    np.testing.assert_allclose(gross_revenue_topside, gross_revenue_botside, atol=1e-6)


def test_consistency_ets():
    # Calculated result Oil
    oil_revenue = (psc._oil_ftp +
                   psc._oil_ic +
                   psc._oil_cost_recovery +
                   psc._oil_ets_before_transfer)

    # Calculated result Gas
    gas_revenue = (psc._gas_ftp +
                   psc._gas_ic +
                   psc._gas_cost_recovery +
                   psc._gas_ets_before_transfer)

    # Expected result Oil
    oil_revenue_base = psc._oil_revenue
    gas_revenue_base = psc._gas_revenue

    np.testing.assert_allclose(oil_revenue, oil_revenue_base)
    np.testing.assert_allclose(gas_revenue, gas_revenue_base)


def test_consolidated_revenue():
    # Calculated result
    engine = psc._consolidated_revenue

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_revenue')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_depreciation():
    # Calculated result
    engine = psc._consolidated_depreciation

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_depreciation')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_non_capital():
    # Calculated result
    engine = psc._consolidated_non_capital

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_non_capital')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_ftp():
    # Calculated result
    engine = psc._consolidated_ftp

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_ftp')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_ftp_ctr():
    # Calculated result
    engine = psc._consolidated_ftp_ctr

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_ftp_ctr')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_ftp_gov():
    # Calculated result
    engine = psc._consolidated_ftp_gov

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_ftp_gov')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_ic():
    # Calculated result
    engine = psc._consolidated_ic

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_ic')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_ic_unrec():
    # Calculated result
    engine = psc._consolidated_ic_unrecovered

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_ic_unrec')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_ic_paid():
    # Calculated result
    engine = psc._consolidated_ic_paid

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_ic_paid')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_unrec_cost():
    # Calculated result
    engine = psc._consolidated_unrecovered_before_transfer

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_unrec_cost')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_cost_recovery_before_transfer():
    # Calculated result
    engine = psc._consolidated_cost_recovery_before_transfer

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_cost_recovery_before_transfer')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_ets_before_transfer():
    # Calculated result
    engine = psc._consolidated_ets_before_transfer

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_ets_before_transfer')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_unrec_after_transfer():
    # Calculated result
    engine = psc._consolidated_unrecovered_after_transfer

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_unrec_after_transfer')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_cost_recovery_after_transfer():
    # Calculated result
    engine = psc._consolidated_cost_recovery_after_tf

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_cost_recovery_after_transfer')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_ets_after_transfer():
    # Calculated result
    engine = psc._consolidated_ets_after_transfer

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_ets_after_transfer')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_ctr_share():
    # Calculated result
    engine = psc._consolidated_contractor_share

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_ctr_share')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_gov_share():
    # Calculated result
    engine = psc._consolidated_government_share

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_gov_share')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_dmo_volume():
    # Calculated result
    engine = psc._consolidated_dmo_volume

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_dmo_volume')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_dmo_fee():
    # Calculated result
    engine = psc._consolidated_dmo_fee

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_dmo_fee')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_ddmo():
    # Calculated result
    engine = psc._consolidated_ddmo

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_ddmo')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_taxable_income():
    # Calculated result
    engine = psc._consolidated_taxable_income

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_taxable_income')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_tax_due():
    # Calculated result
    engine = psc._consolidated_tax_due

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_tax_due')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_unpaid_tax_balance():
    # Calculated result
    engine = psc._consolidated_unpaid_tax_balance

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_unpaid_tax_balance')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_tax_payment():
    # Calculated result
    engine = psc._consolidated_tax_payment

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_tax_payment')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_ctr_net_share():
    # Calculated result
    engine = psc._consolidated_ctr_net_share

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_ctr_net_share')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_gov_take():
    # Calculated result
    engine = psc._consolidated_government_take

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_gov_take')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)


def test_consolidated_cashflow():
    # Calculated result
    engine = psc._consolidated_cashflow

    # Expected result
    base = load_testing(dataset_type='case3', key='cnsltd_cashflow')

    # Execute testing
    np.testing.assert_allclose(engine, base, atol=1e-6)






"""
This test code will be changed in the version 1_2_0
"""
# ToDo: Rewrite the test code in the version 1.2.0

import numpy as np
from pyscnomics.dataset.sample import load_data, load_testing


psc = load_data(dataset_type='small_gas', contract_type='cost_recovery')
psc.run()


def test_revenue():
    """
    Test module to test value of revenue calculated by pyscnomics.
    """

    revenue = psc._gas_revenue
    revenue_base = np.asarray(load_testing(dataset_type='small_gas', key='Revenue'))

    np.testing.assert_allclose(revenue, revenue_base)


def test_depreciation():
    """
    Test module for testing the Depreciation.
    """

    gas_depreciation = psc._gas_depreciation
    gas_depreciation_base = load_testing(dataset_type='small_gas', key='Depreciation')

    np.testing.assert_allclose(gas_depreciation, gas_depreciation_base)


def test_non_capital():

    gas_non_capital = psc._gas_non_capital
    gas_non_capital_base = load_testing(dataset_type='small_gas', key='Non Capital')

    np.testing.assert_allclose(gas_non_capital, gas_non_capital_base)


def test_ftp():

    gas_ftp = psc._gas_ftp
    gas_ftp_base = load_testing(dataset_type='small_gas', key='FTP')

    np.testing.assert_allclose(gas_ftp, gas_ftp_base)

def test_ftp_ctr():

    gas_ftp_ctr = psc._gas_ftp_ctr
    gas_ftp_ctr_base = load_testing(dataset_type='small_gas', key='FTP CTR')

    np.testing.assert_allclose(gas_ftp_ctr, gas_ftp_ctr_base)

def test_ftp_gov():

    gas_ftp_gov = psc._gas_ftp_gov
    gas_ftp_gov_base = load_testing(dataset_type='small_gas', key='FTP GOV')

    np.testing.assert_allclose(gas_ftp_gov, gas_ftp_gov_base)

def test_ic():
    """
    Test module for testing the Investment Credit (IC).
    """

    gas_ic = psc._gas_ic
    gas_ic_unrecovered = psc._gas_ic_unrecovered
    gas_ic_paid = psc._gas_ic_paid

    gas_ic_base = load_testing(dataset_type='small_gas', key='IC')
    gas_ic_unrecovered_base = load_testing(dataset_type='small_gas', key='IC Unrec')
    gas_ic_paid_base = load_testing(dataset_type='small_gas', key='IC Paid')

    np.testing.assert_allclose(gas_ic, gas_ic_base)
    np.testing.assert_allclose(gas_ic_unrecovered, gas_ic_unrecovered_base)
    np.testing.assert_allclose(gas_ic_paid, gas_ic_paid_base)


def test_unrec_before_transfer():
    """
    Test module for testing the Unrecovered Cost Before Transfer.
    """

    unrec_before_transfer = psc._gas_unrecovered_before_transfer
    unrec_before_transfer_base = load_testing(dataset_type='small_gas', key='Unrec Cost')

    np.testing.assert_allclose(unrec_before_transfer, unrec_before_transfer_base)


def test_cost_tobe_recovered():
    """
    Test module for testing the Cost To Be Recover.
    """

    gas_cost_to_be_recovered = psc._gas_cost_to_be_recovered
    gas_cost_to_be_recovered_base = load_testing(dataset_type='small_gas', key='Cost To Be Recovered')

    np.testing.assert_allclose(gas_cost_to_be_recovered, gas_cost_to_be_recovered_base)


def test_cost_recovery():

    # Calculated result
    gas_cost_recovery = psc._gas_cost_recovery

    # Expected result
    gas_cost_recovery_base = load_testing(dataset_type='small_gas', key='Cost Recovery')

    # Execute testing
    np.testing.assert_allclose(gas_cost_recovery, gas_cost_recovery_base)


def test_ets_before_transfer():

    # Calculated result
    gas_ets_before_transfer = psc._gas_ets_before_transfer

    # Expected result
    gas_ets_before_transfer_base = load_testing(dataset_type='small_gas', key='ETS Before Transfer')

    # Execute testing
    np.testing.assert_allclose(gas_ets_before_transfer, gas_ets_before_transfer_base)


def test_transfer():

    # Calculated result
    transfer_to_oil = psc._transfer_to_oil
    transfer_to_gas = psc._transfer_to_gas

    # Expected result
    transfer_to_oil_base = load_testing(dataset_type='small_gas', key='Transfer to Oil')
    transfer_to_gas_base = load_testing(dataset_type='small_gas', key='Transfer to Gas')

    # Execute testing
    np.testing.assert_allclose(transfer_to_oil, transfer_to_oil_base)
    np.testing.assert_allclose(transfer_to_gas, transfer_to_gas_base)


def test_unrecovered_after_transfer():

    # Calculated result
    gas_unrecovered_after_transfer = psc._gas_unrecovered_after_transfer

    # Expected result
    gas_unrecovered_after_transfer_base = load_testing(dataset_type='small_gas', key='Unrec After Transfer')

    # Execute testing
    np.testing.assert_allclose(gas_unrecovered_after_transfer, gas_unrecovered_after_transfer_base)


def test_ets_after_transfer():

    # Calculated result
    gas_ets_after_transfer = psc._gas_ets_after_transfer

    # Expected result
    gas_ets_after_transfer_base = load_testing(dataset_type='small_gas', key='ETS After Transfer')

    # Execute testing
    np.testing.assert_allclose(gas_ets_after_transfer, gas_ets_after_transfer_base)


def test_equity_share():

    # Calculated result
    gas_contractor_share = psc._gas_contractor_share
    gas_government_share = psc._gas_government_share

    # Expected result
    gas_contractor_share_base = load_testing(dataset_type='small_gas', key='Contractor Share')
    gas_government_share_base = load_testing(dataset_type='small_gas', key='Government Share')

    # Execute testing
    np.testing.assert_allclose(gas_contractor_share_base, gas_contractor_share)
    np.testing.assert_allclose(gas_government_share, gas_government_share_base)


def test_dmo():

    # Calculated result
    gas_dmo_volume = psc._gas_dmo_volume
    gas_dmo_fee = psc._gas_dmo_fee
    gas_ddmo = psc._gas_ddmo

    # Expected result
    gas_dmo_volume_base = load_testing(dataset_type='small_gas', key='DMO')
    gas_dmo_fee_base = load_testing(dataset_type='small_gas', key='DMO Fee')
    gas_ddmo_base = load_testing(dataset_type='small_gas', key='Net DMO')

    # Execute testing
    np.testing.assert_allclose(gas_dmo_volume, gas_dmo_volume_base)
    np.testing.assert_allclose(gas_dmo_fee, gas_dmo_fee_base)
    np.testing.assert_allclose(gas_ddmo, gas_ddmo_base)


def test_taxable_income():

    # Calculated result
    gas_taxable_income = psc._gas_taxable_income

    # Expected result
    gas_taxable_income_base = load_testing(dataset_type='small_gas', key='Taxable Income')

    # Execute testing
    np.testing.assert_allclose(gas_taxable_income, gas_taxable_income_base)


def test_consistency():
    # Top side result
    gross_revenue_topside = psc._oil_revenue + psc._gas_revenue

    # Bottom side result
    gross_revenue_botside = psc._oil_contractor_take + psc._gas_contractor_take + psc._oil_government_take + psc._gas_government_take

    # Execute testing
    np.testing.assert_allclose(gross_revenue_topside, gross_revenue_botside, rtol=1e-6)


def test_consistency_ets():
    # Calculated result Oil
    oil_revenue = (psc._oil_ftp +
                   psc._oil_ic +
                   psc._oil_cost_recovery +
                   psc._oil_ets_before_transfer)

    # Calculated result Gas
    gas_revenue = (psc._gas_ftp +
                   psc._gas_ic +
                   psc._gas_cost_recovery +
                   psc._gas_ets_before_transfer)

    # Expected result Oil
    oil_revenue_base = psc._oil_revenue
    gas_revenue_base = psc._gas_revenue

    np.testing.assert_allclose(oil_revenue, oil_revenue_base)
    np.testing.assert_allclose(gas_revenue, gas_revenue_base)


"""
Handles calculations associated with PSC Gross Split.
"""
import warnings
from dataclasses import dataclass, field
import numpy as np

from pyscnomics.contracts.project import BaseProject
from pyscnomics.contracts import psc_tools
from pyscnomics.econ.selection import (FluidType,
                                       GrossSplitRegime,
                                       TaxRegime,
                                       TaxType,
                                       DeprMethod,
                                       OtherRevenue,
                                       InflationAppliedTo,
                                       VariableSplit522017,
                                       VariableSplit082017)
from pyscnomics.econ.results import CashFlow
from pyscnomics.econ.depreciation import unit_of_production_rate


class GrossSplitException(Exception):
    """Exception to raise for a misuse of GrossSplit class"""

    pass


class SunkCostException(Exception):
    """Exception to raise for a misuse of Sunk Cost Method"""

    pass


class CumulativeProductionSplitException(Exception):
    """Exception to raise for a misuse of Cumulative Production Split Offset"""

    pass


@dataclass
class GrossSplit(BaseProject):
    """
    Dataclass that represents Gross Split (GS) contract.

    Parameters
    ----------
    field_status: str | VariableSplit522017.FieldStatus | VariableSplit082017.FieldStatus
        The field status of the corresponding contract.
    field_loc: str | VariableSplit522017.FieldLocation | VariableSplit082017.FieldLocation
        The field location of the corresponding contract.
    res_depth: str | VariableSplit522017.ReservoirDepth | VariableSplit082017.ReservoirDepth
        The reservoir depth of the corresponding contract
    infra_avail: str | VariableSplit522017.InfrastructureAvailability | VariableSplit082017.InfrastructureAvailability
        The infrastructure availability of the corresponding contract.
    res_type: str | VariableSplit522017.ReservoirType | VariableSplit082017.ReservoirType
        The reservoir type of the corresponding contract.
    api_oil: str | VariableSplit522017.APIOil | VariableSplit082017.APIOil
        The Oil's API of the corresponding contract.
    domestic_use: str | VariableSplit522017.DomesticUse | VariableSplit082017.DomesticUse
        The domestic use or local content of the corresponding contract.
    prod_stage: str | VariableSplit522017.ProductionStage | VariableSplit082017.ProductionStage
        The production stage of the corresponding contract.
    co2_content: str | VariableSplit522017.CO2Content | VariableSplit082017.CO2Content
        The CO2 content of the corresponding contract.
    h2s_content: str | VariableSplit522017.H2SContent | VariableSplit082017.H2SContent
        The H2S content of the corresponding contract.
    base_split_ctr_oil: float
        The contractor base split for oil of the corresponding contract.
    base_split_ctr_gas: float
        The contractor base split for gas of the corresponding contract.
    split_ministry_disc: float
        The ministerial discretion split of the corresponding contract.
    oil_dmo_volume_portion: float
        The Oil's Domestic Market Obligation (DMO) volume portion.
    oil_dmo_fee_portion: float
        The Oil's DMO fee portion.
    oil_dmo_holiday_duration: int
        The duration of the Oil DMO Holiday in month unit.
    gas_dmo_volume_portion: float
        The Gas's Domestic Market Obligation (DMO) volume portion.
    gas_dmo_fee_portion: float
        The Gas's DMO fee portion.
    gas_dmo_holiday_duration: int
        The duration of the Gas's DMO Holiday in month unit.
    """
    field_status: str | VariableSplit522017.FieldStatus | VariableSplit082017.FieldStatus = field(default='No POD')
    field_loc: str | VariableSplit522017.FieldLocation | VariableSplit082017.FieldLocation = field(default='Onshore')
    res_depth: str | VariableSplit522017.ReservoirDepth | VariableSplit082017.ReservoirDepth = field(default='<=2500')
    infra_avail: str | VariableSplit522017.InfrastructureAvailability | VariableSplit082017.InfrastructureAvailability = field(default='Well Developed')
    res_type: str | VariableSplit522017.ReservoirType | VariableSplit082017.ReservoirType = field(default='Conventional')
    api_oil: str | VariableSplit522017.APIOil | VariableSplit082017.APIOil = field(default='<25')
    domestic_use: str | VariableSplit522017.DomesticUse | VariableSplit082017.DomesticUse = field(default='50<=x<70')
    prod_stage: str | VariableSplit522017.ProductionStage | VariableSplit082017.ProductionStage = field(default='Secondary')
    co2_content: str | VariableSplit522017.CO2Content | VariableSplit082017.CO2Content = field(default='<5')
    h2s_content: str | VariableSplit522017.H2SContent | VariableSplit082017.H2SContent = field(default='<100')

    base_split_ctr_oil: float = field(default=0.43)
    base_split_ctr_gas: float = field(default=0.48)
    split_ministry_disc: float = field(default=0.08)

    oil_dmo_volume_portion: float = field(default=0.25)
    oil_dmo_fee_portion: float = field(default=1.0)
    oil_dmo_holiday_duration: int = field(default=60)

    gas_dmo_volume_portion: float = field(default=1.0)
    gas_dmo_fee_portion: float = field(default=1.0)
    gas_dmo_holiday_duration: int = field(default=60)

    conversion_boe_to_scf: float = field(default=5.615, init=False, repr=False)

    _oil_depreciation: np.ndarray = field(default=None, init=False, repr=False)
    _gas_depreciation: np.ndarray = field(default=None, init=False, repr=False)
    _oil_undepreciated_asset: np.ndarray = field(default=None, init=False, repr=False)
    _gas_undepreciated_asset: np.ndarray = field(default=None, init=False, repr=False)
    _oil_non_capital: np.ndarray = field(default=None, init=False, repr=False)
    _gas_non_capital: np.ndarray = field(default=None, init=False, repr=False)

    _cumulative_prod: np.ndarray = field(default=None, init=False, repr=False)

    _oil_base_split: np.ndarray = field(default=None, init=False, repr=False)
    _gas_base_split: np.ndarray = field(default=None, init=False, repr=False)
    _variable_split: float = field(default=0, init=False, repr=False)
    _var_split_array: np.ndarray = field(default=None, init=False, repr=False)
    _oil_prog_split: np.ndarray = field(default=None, init=False, repr=False)
    _gas_prog_split: np.ndarray = field(default=None, init=False, repr=False)
    _oil_ctr_split: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ctr_split: np.ndarray = field(default=None, init=False, repr=False)
    _oil_ctr_share_before_transfer: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ctr_share_before_transfer: np.ndarray = field(default=None, init=False, repr=False)
    _oil_gov_share: np.ndarray = field(default=None, init=False, repr=False)
    _gas_gov_share: np.ndarray = field(default=None, init=False, repr=False)
    _oil_total_expenses: np.ndarray = field(default=None, init=False, repr=False)
    _gas_total_expenses: np.ndarray = field(default=None, init=False, repr=False)

    _oil_cost_tobe_deducted: np.ndarray = field(default=None, init=False, repr=False)
    _gas_cost_tobe_deducted: np.ndarray = field(default=None, init=False, repr=False)
    _oil_carward_deduct_cost: np.ndarray = field(default=None, init=False, repr=False)
    _gas_carward_deduct_cost: np.ndarray = field(default=None, init=False, repr=False)
    _oil_deductible_cost: np.ndarray = field(default=None, init=False, repr=False)
    _gas_deductible_cost: np.ndarray = field(default=None, init=False, repr=False)

    _transfer_to_oil: np.ndarray = field(default=None, init=False, repr=False)
    _transfer_to_gas: np.ndarray = field(default=None, init=False, repr=False)
    _oil_carward_cost_aftertf: np.ndarray = field(default=None, init=False, repr=False)
    _gas_carward_cost_aftertf: np.ndarray = field(default=None, init=False, repr=False)

    _oil_ctr_share_after_transfer: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ctr_share_after_transfer: np.ndarray = field(default=None, init=False, repr=False)

    _oil_net_operating_profit: np.ndarray = field(default=None, init=False, repr=False)
    _gas_net_operating_profit: np.ndarray = field(default=None, init=False, repr=False)

    _oil_dmo_volume: np.ndarray = field(default=None, init=False, repr=False)
    _oil_dmo_fee: np.ndarray = field(default=None, init=False, repr=False)
    _oil_ddmo: np.ndarray = field(default=None, init=False, repr=False)

    _gas_dmo_volume: np.ndarray = field(default=None, init=False, repr=False)
    _gas_dmo_fee: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ddmo: np.ndarray = field(default=None, init=False, repr=False)

    _oil_taxable_income: np.ndarray = field(default=None, init=False, repr=False)
    _gas_taxable_income: np.ndarray = field(default=None, init=False, repr=False)

    _tax_rate_arr: np.ndarray = field(default=None, init=False, repr=False)

    _oil_tax: np.ndarray = field(default=None, init=False, repr=False)
    _gas_tax: np.ndarray = field(default=None, init=False, repr=False)

    _oil_ctr_net_share: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ctr_net_share: np.ndarray = field(default=None, init=False, repr=False)

    _oil_ctr_cashflow: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ctr_cashflow: np.ndarray = field(default=None, init=False, repr=False)

    _oil_government_take: np.ndarray = field(default=None, init=False, repr=False)
    _gas_government_take: np.ndarray = field(default=None, init=False, repr=False)

    _oil_cashflow: CashFlow = field(default=None, init=False, repr=False)
    _gas_cashflow: CashFlow = field(default=None, init=False, repr=False)

    # Consolidated Attributes
    _consolidated_revenue: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_capital_cost: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_intangible: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_sunk_cost: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_opex: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_asr: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_non_capital: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_depreciation: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_undepreciated_asset: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_ctr_share_before_tf: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_gov_share_before_tf: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_total_expenses: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_cost_tobe_deducted: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_carward_deduct_cost: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_deductible_cost: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_carward_cost_aftertf: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_ctr_share_after_transfer: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_net_operating_profit: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_dmo_volume: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_dmo_fee: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_ddmo: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_taxable_income: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_tax_payment: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_ctr_net_share: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_cashflow: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_government_take: np.ndarray = field(default=None, init=False, repr=False)

    # Attributes fof containing the 100% contractor split warning
    _oil_ctr_split_prior_bracket: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ctr_split_prior_bracket: np.ndarray = field(default=None, init=False, repr=False)
    _oil_year_maximum_ctr_split: np.ndarray = field(default=None, init=False, repr=False)
    _gas_year_maximum_ctr_split: np.ndarray = field(default=None, init=False, repr=False)


    def _check_attributes(self):
        """
        Function to check the Gross Split input.
        -------

        """
        # Defining any attributes that in the form of fraction
        fraction_attributes = (
            ('base_split_ctr_oil', self.base_split_ctr_oil),
            ('base_split_ctr_gas', self.base_split_ctr_gas),
            ('split_ministry_disc', self.split_ministry_disc),
            ('oil_dmo_volume_portion', self.oil_dmo_volume_portion),
            ('oil_dmo_fee_portion', self.oil_dmo_fee_portion),
            ('gas_dmo_volume_portion', self.gas_dmo_volume_portion),
            ('gas_dmo_fee_portion', self.gas_dmo_fee_portion)
        )

        for attr_name, attr in fraction_attributes:
            if attr > 1.0 or attr < 0.0:
                range_type = 'exceeding 1.0' if attr > 1.0 else 'below 0.0'
                raise GrossSplitException(
                    f"The {attr_name} value, {attr}, is {range_type}. "
                    f"The allowed range for this attribute is between 0.0 and 1.0."
                )

        discrete_attributes = (
            ('oil_dmo_holiday_duration', self.oil_dmo_holiday_duration),
            ('gas_dmo_holiday_duration', self.gas_dmo_holiday_duration)
        )

        for attr_name, attr in discrete_attributes:
            if attr < 0:
                raise GrossSplitException(
                    f"The {attr_name} value, {attr}, is below 0. "
                    f"The minimum value for this attribute is 0."
                )

    def _wrapper_variable_split(self,
                                regime: GrossSplitRegime = GrossSplitRegime.PERMEN_ESDM_20_2019):
        """
        Function to wrap the variable split function.

        Parameters
        ----------
        regime: GrossSplitRegime
         The selection of the Gross Split Regime

        Returns
        -------
        variable_split_func
            The function of the variable split

        Notes
        -------
        (1) Gross Split Regime PERMEN ESDM No. 52 Tahun 2017, PERMEN ESDM No. 20 Tahun 2019,
        and PERMEN ESDM No. 12 Tahun 2020 are having the same variable split value. The complete differences
        could be seen in official documents.
        """

        if regime == GrossSplitRegime.PERMEN_ESDM_8_2017:
            variable_split_func = self._get_var_split_08_2017()

        elif (regime == GrossSplitRegime.PERMEN_ESDM_52_2017 or
              GrossSplitRegime.PERMEN_ESDM_20_2019 or
              GrossSplitRegime.PERMEN_ESDM_12_2020):
            variable_split_func = self._get_var_split_52_2017()

        else:
            raise GrossSplitException(
                f"The Gross Split regime, {regime}, is not recognized"
            )

        return variable_split_func

    def _get_var_split_08_2017(self):
        """
        A function to get the value of Variable Split based on the given parameters.

        Returns
        -------
        _variable_split: float
            The value of variable split.
        """
        # The string parameters is being keep to backward compatibility
        params_str = {
            'field_status': {
                'POD I': 0.05,
                'POD II': 0.0,
                'POFD': 0.0,
                'No POD': -0.05,
            },
            'field_loc': {
                'Onshore': 0.0,
                'Offshore (0<h<=20)': 0.08,
                'Offshore (20<h<=50)': 0.1,
                'Offshore (50<h<=150)': 0.12,
                'Offshore (150<h<=1000)': 0.14,
                'Offshore (h>1000)': 0.16,
            },
            'res_depth': {
                '<=2500': 0.0,
                '>2500': 0.01,
            },
            'infra_avail': {
                'Well Developed': 0.0,
                'New Frontier': 0.02,
            },
            'res_type': {
                'Conventional': 0.0,
                'Non Conventional': 0.16,
            },
            'co2_content': {
                '<5': 0.0,
                '5<=x<10': 0.005,
                '10<=x<20': 0.01,
                '20<=x<40': 0.015,
                '40<=x<60': 0.02,
                'x>=60': 0.04,
            },
            'h2s_content': {
                '<100': 0.0,
                '100<=x<300': 0.005,
                '300<=x<500': 0.0075,
                'x>=500': 0.01,
            },
            'api_oil': {
                '<25': 0.01,
                '>=25': 0.0,
            },
            'domestic_use': {
                '<30': 0.0,
                '30<=x<50': 0.02,
                '50<=x<70': 0.03,
                '70<=x<100': 0.04,
            },
            'prod_stage': {
                'Primary': 0.0,
                'Secondary': 0.03,
                'Tertiary': 0.05,
            }
        }

        params_enum = {
            'field_status': {
                VariableSplit082017.FieldStatus.POD_I: 0.05,
                VariableSplit082017.FieldStatus.POD_II: 0.0,
                VariableSplit082017.FieldStatus.POFD: 0.0,
                VariableSplit082017.FieldStatus.NO_POD: -0.05,
            },
            'field_loc': {
                VariableSplit082017.FieldLocation.ONSHORE: 0.0,
                VariableSplit082017.FieldLocation.OFFSHORE_0_UNTIL_LESSEQUAL_20: 0.08,
                VariableSplit082017.FieldLocation.OFFSHORE_20_UNTIL_LESSEQUAL_50: 0.1,
                VariableSplit082017.FieldLocation.OFFSHORE_50_UNTIL_LESSEQUAL_150: 0.12,
                VariableSplit082017.FieldLocation.OFFSHORE_150_UNTIL_LESSEQUAL_1000: 0.14,
                VariableSplit082017.FieldLocation.OFFSHORE_GREATERTHAN_1000: 0.16,
            },
            'res_depth': {
                VariableSplit082017.ReservoirDepth.LESSEQUAL_2500: 0.0,
                VariableSplit082017.ReservoirDepth.GREATERTHAN_2500: 0.01,
            },
            'infra_avail': {
                VariableSplit082017.InfrastructureAvailability.WELL_DEVELOPED: 0.0,
                VariableSplit082017.InfrastructureAvailability.NEW_FRONTIER: 0.02,
            },
            'res_type': {
                VariableSplit082017.ReservoirType.CONVENTIONAL: 0.0,
                VariableSplit082017.ReservoirType.NON_CONVENTIONAL: 0.16,
            },
            'co2_content': {
                VariableSplit082017.CO2Content.LESSTHAN_5: 0.0,
                VariableSplit082017.CO2Content.EQUAL_5_UNTIL_LESSTHAN_10: 0.005,
                VariableSplit082017.CO2Content.EQUAL_10_UNTIL_LESSTHAN_20: 0.01,
                VariableSplit082017.CO2Content.EQUAL_20_UNTIL_LESSTHAN_40: 0.015,
                VariableSplit082017.CO2Content.EQUAL_40_UNTIL_LESSTHAN_60: 0.02,
                VariableSplit082017.CO2Content.EQUALGREATERTHAN_60: 0.04,
            },
            'h2s_content': {
                VariableSplit082017.H2SContent.LESSTHAN_100: 0.0,
                VariableSplit082017.H2SContent.EQUAL_100_UNTIL_LESSTHAN_300: 0.005,
                VariableSplit082017.H2SContent.EQUAL_300_UNTIL_LESSTHAN_500: 0.0075,
                VariableSplit082017.H2SContent.EQUALGREATERTHAN_500: 0.01,
            },
            'api_oil': {
                VariableSplit082017.APIOil.LESSTHAN_25: 0.01,
                VariableSplit082017.APIOil.EQUALGREATERTHAN_25: 0.0,
            },
            'domestic_use': {
                VariableSplit082017.DomesticUse.LESSTHAN_30: 0.0,
                VariableSplit082017.DomesticUse.EQUAL_30_UNTIL_LESSTHAN_50: 0.02,
                VariableSplit082017.DomesticUse.EQUAL_50_UNTIL_LESSTHAN_70: 0.03,
                VariableSplit082017.DomesticUse.EQUAL_70_UNTIL_LESSTHAN_100: 0.04,
            },
            'prod_stage': {
                VariableSplit082017.ProductionStage.PRIMARY: 0.0,
                VariableSplit082017.ProductionStage.SECONDARY: 0.03,
                VariableSplit082017.ProductionStage.TERTIARY: 0.05,
            }
        }
        source_dict = {
            'field_status': self.field_status,
            'field_loc': self.field_loc,
            'res_depth': self.res_depth,
            'infra_avail': self.infra_avail,
            'res_type': self.res_type,
            'api_oil': self.api_oil,
            'domestic_use': self.domestic_use,
            'prod_stage': self.prod_stage,
            'co2_content': self.co2_content,
            'h2s_content': self.h2s_content,
        }

        variable_split = np.array([
            params_str[key][param] if isinstance(param, str) else params_enum[key][param]
            for key, param in source_dict.items()
        ], dtype=float)

        self._variable_split = float(np.sum(variable_split))

    def _get_var_split_52_2017(self):
        """
        A function to get the value of Variable Split based on the given parameters.

        Returns
        -------
        _variable_split: float
            The value of variable split.
        """
        params_str = {
            'field_status': {
                'POD I': 0.05,
                'POD II': 0.03,
                'No POD': 0,
            },
            'field_loc': {
                'Onshore': 0,
                'Offshore (0<h<=20)': 0.08,
                'Offshore (20<h<=50)': 0.1,
                'Offshore (50<h<=150)': 0.12,
                'Offshore (150<h<=1000)': 0.14,
                'Offshore (h>1000)': 0.16,
            },
            'res_depth': {
                '<=2500': 0,
                '>2500': 0.01,
            },
            'infra_avail': {
                'Well Developed': 0,
                'New Frontier Offshore': 0.02,
                'New Frontier Onshore': 0.04,
            },
            'res_type': {
                'Conventional': 0,
                'Non Conventional': 0.16,
            },
            'co2_content': {
                '<5': 0,
                '5<=x<10': 0.005,
                '10<=x<20': 0.01,
                '20<=x<40': 0.015,
                '40<=x<60': 0.02,
                'x>=60': 0.04,
            },
            'h2s_content': {
                '<100': 0,
                '100<=x<1000': 0.01,
                '1000<=x<2000': 0.02,
                '2000<=x<3000': 0.03,
                '3000<=x<4000': 0.04,
                'x>=4000': 0.05,
            },
            'api_oil': {
                '<25': 0.01,
                '>=25': 0,
            },
            'domestic_use': {
                '30<=x<50': 0.02,
                '50<=x<70': 0.03,
                '70<=x<100': 0.04,
            },
            'prod_stage': {
                'Primary': 0,
                'Secondary': 0.06,
                'Tertiary': 0.1,
            }
        }

        params_enum = {
            'field_status': {
                VariableSplit522017.FieldStatus.POD_I: 0.05,
                VariableSplit522017.FieldStatus.POD_II: 0.03,
                VariableSplit522017.FieldStatus.NO_POD: 0.0,
            },
            'field_loc': {
                VariableSplit522017.FieldLocation.ONSHORE: 0.0,
                VariableSplit522017.FieldLocation.OFFSHORE_0_UNTIL_LESSEQUAL_20: 0.08,
                VariableSplit522017.FieldLocation.OFFSHORE_20_UNTIL_LESSEQUAL_50: 0.1,
                VariableSplit522017.FieldLocation.OFFSHORE_50_UNTIL_LESSEQUAL_150: 0.12,
                VariableSplit522017.FieldLocation.OFFSHORE_150_UNTIL_LESSEQUAL_1000: 0.14,
                VariableSplit522017.FieldLocation.OFFSHORE_GREATERTHAN_1000: 0.16,
            },
            'res_depth': {
                VariableSplit522017.ReservoirDepth.LESSEQUAL_2500: 0.0,
                VariableSplit522017.ReservoirDepth.GREATERTHAN_2500: 0.01,
            },
            'infra_avail': {
                VariableSplit522017.InfrastructureAvailability.WELL_DEVELOPED: 0.0,
                VariableSplit522017.InfrastructureAvailability.NEW_FRONTIER_OFFSHORE: 0.02,
                VariableSplit522017.InfrastructureAvailability.NEW_FRONTIER_ONSHORE: 0.04,
            },
            'res_type': {
                VariableSplit522017.ReservoirType.CONVENTIONAL: 0.0,
                VariableSplit522017.ReservoirType.NON_CONVENTIONAL: 0.16,
            },
            'co2_content': {
                VariableSplit522017.CO2Content.LESSTHAN_5: 0.0,
                VariableSplit522017.CO2Content.EQUAL_5_UNTIL_LESSTHAN_10: 0.005,
                VariableSplit522017.CO2Content.EQUAL_10_UNTIL_LESSTHAN_20: 0.01,
                VariableSplit522017.CO2Content.EQUAL_20_UNTIL_LESSTHAN_40: 0.015,
                VariableSplit522017.CO2Content.EQUAL_40_UNTIL_LESSTHAN_60: 0.02,
                VariableSplit522017.CO2Content.EQUALGREATERTHAN_60: 0.04,
            },
            'h2s_content': {
                VariableSplit522017.H2SContent.LESSTHAN_100: 0.0,
                VariableSplit522017.H2SContent.EQUAL_100_UNTIL_LESSTHAN_1000: 0.01,
                VariableSplit522017.H2SContent.EQUAL_1000_UNTIL_LESSTHAN_2000: 0.02,
                VariableSplit522017.H2SContent.EQUAL_2000_UNTIL_LESSTHAN_3000: 0.03,
                VariableSplit522017.H2SContent.EQUAL_3000_UNTIL_LESSTHAN_4000: 0.04,
                VariableSplit522017.H2SContent.EQUALGREATERTHAN_4000: 0.05,
            },
            'api_oil': {
                VariableSplit522017.APIOil.LESSTHAN_25: 0.01,
                VariableSplit522017.APIOil.EQUALGREATERTHAN_25: 0.0,
            },
            'domestic_use': {
                VariableSplit522017.DomesticUse.EQUAL_30_UNTIL_LESSTHAN_50: 0.02,
                VariableSplit522017.DomesticUse.EQUAL_50_UNTIL_LESSTHAN_70: 0.03,
                VariableSplit522017.DomesticUse.EQUAL_70_UNTIL_LESSTHAN_100: 0.04,
            },
            'prod_stage': {
                VariableSplit522017.ProductionStage.PRIMARY: 0.0,
                VariableSplit522017.ProductionStage.SECONDARY: 0.06,
                VariableSplit522017.ProductionStage.TERTIARY: 0.1,
            }
        }

        source_dict = {
            'field_status': self.field_status,
            'field_loc': self.field_loc,
            'res_depth': self.res_depth,
            'infra_avail': self.infra_avail,
            'res_type': self.res_type,
            'api_oil': self.api_oil,
            'domestic_use': self.domestic_use,
            'prod_stage': self.prod_stage,
            'co2_content': self.co2_content,
            'h2s_content': self.h2s_content,
        }

        variable_split = np.array([
            params_str[key][param] if isinstance(param, str) else params_enum[key][param]
            for key, param in source_dict.items()
        ], dtype=float)

        self._variable_split = float(np.sum(variable_split))

    def _wrapper_progressive_split(
            self,
            fluid: FluidType,
            price: np.ndarray,
            cum: np.ndarray,
            regime: GrossSplitRegime = GrossSplitRegime.PERMEN_ESDM_20_2019):

        if (regime == GrossSplitRegime.PERMEN_ESDM_52_2017 or
                regime == GrossSplitRegime.PERMEN_ESDM_20_2019 or
                regime == GrossSplitRegime.PERMEN_ESDM_12_2020):
            prog_price_split = self._get_prog_price_split_52_2017(fluid, price)
            prog_cum_split = self._get_prog_cum_split_52_2017(cum)

        elif regime == GrossSplitRegime.PERMEN_ESDM_8_2017:
            prog_price_split = self._get_prog_price_split_08_2017(fluid, price)
            prog_cum_split = self._get_prog_cum_split_08_2017(cum)

        else:
            prog_price_split = ValueError('Not Recognized Gross Split Regime')
            prog_cum_split = ValueError('Not Recognized Gross Split Regime')

        return prog_price_split, prog_cum_split

    @staticmethod
    def _get_prog_price_split_08_2017(
            fluid: FluidType,
            price: np.ndarray
    ):
        # Indonesia's Ministry Regulations No.08 The Year of 2017. At Appendix B Progressive Component
        if fluid == FluidType.OIL:
            if price < 40:
                ps = 0.075
            elif 40 <= price < 55:
                ps = 0.05
            elif 55 <= price < 70:
                ps = 0.025
            elif 70 <= price < 85:
                ps = 0.0
            elif 85 <= price < 100:
                ps = -0.025
            elif 100 <= price < 115:
                ps = -0.05
            elif 115 <= price:
                ps = -0.075
            else:
                ps = 0
        else:
            ps = 0

        return ps

    @staticmethod
    def _get_prog_price_split_52_2017(
            fluid: FluidType,
            price: np.ndarray
    ):
        # Indonesia's Ministry Regulations No.52 The Year of 2017. At Appendix B Progressive Component
        if fluid == FluidType.OIL:
            ps = np.where(price > 0,
                          (85 - price) * 0.25 / 100,
                          0)

        elif fluid == FluidType.GAS:
            if price < 7:
                ps = (7 - price) * 2.5 / 100
            elif (7 < price) and (price < 10):
                ps = 0
            elif price > 10:
                ps = (10 - price) * 2.5 / 100
            else:
                ps = 0
        else:
            raise ValueError('Unknown fluid type')

        return ps

    @staticmethod
    def _get_prog_cum_split_08_2017(
            cum: np.ndarray | None,
    ):
        # Cumulative Progressive Split
        if 0 < cum < 1000:
            px = 0.05
        elif 1000 <= cum < 10000:
            px = 0.04
        elif 10000 <= cum < 20000:
            px = 0.03
        elif 20000 <= cum < 50000:
            px = 0.02
        elif 50000 <= cum < 150000:
            px = 0.01
        elif 150000 <= cum:
            px = 0
        elif cum is None:
            px = 0
        else:
            raise ValueError('No Regulation exist regarding the cumulative value')

        return px

    @staticmethod
    def _get_prog_cum_split_52_2017(
            cum: np.ndarray | None
    ):
        # Cumulative Progressive Split
        if np.logical_and(np.greater_equal(cum, 0), np.less(cum, 30000)):
            px = 0.1
        elif np.logical_and(np.greater_equal(cum, 30000), np.less(cum, 60000)):
            px = 0.09
        elif np.logical_and(np.greater_equal(cum, 60000), np.less(cum, 90000)):
            px = 0.08
        elif np.logical_and(np.greater_equal(cum, 90000), np.less(cum, 125000)):
            px = 0.06
        elif np.logical_and(np.greater_equal(cum, 125000), np.less(cum, 175000)):
            px = 0.04
        elif np.greater_equal(cum, 175000):
            px = 0
        elif cum is None:
            px = 0
        else:
            raise ValueError('No Regulation exist regarding the cumulative value')

        return px

    @staticmethod
    def _get_deductible_cost(ctr_gross_share, cost_tobe_deducted, carward_deduct_cost):

        carward_deduct_cost = np.concatenate((np.zeros(1), carward_deduct_cost[:-1]))

        return np.where(ctr_gross_share > (cost_tobe_deducted + carward_deduct_cost),
                        cost_tobe_deducted + carward_deduct_cost,
                        ctr_gross_share)

    def _get_sunk_cost(self, sunk_cost_reference_year: int):
        oil_cost_raw = (
                self._oil_capital_expenditures
                + self._oil_non_capital
        )
        self._oil_sunk_cost = oil_cost_raw[
                              : (sunk_cost_reference_year - self.start_date.year + 1)
                              ]

        gas_cost_raw = (
                self._gas_capital_expenditures
                + self._gas_non_capital
        )
        self._gas_sunk_cost = gas_cost_raw[
                              : (sunk_cost_reference_year - self.start_date.year + 1)
                              ]

        if sunk_cost_reference_year == self.start_date.year:
            self._oil_sunk_cost = np.zeros(1)
            self._gas_sunk_cost = np.zeros(1)

    def _get_year_maximum_split(
            self,
            ctr_split: np.ndarray,
            fluid: str
    ):
        """
        Function to get the years of when the contractor have maximum split 100% or more.

        Parameters
        ----------
        ctr_split: np.ndarray
            The array of the contractor split
        fluid: str
            The fluid type that the contractor split being observed.

        Returns
        -------
        out: np.ndarray
            The array of the years when the contractor have the maximum split 100%.

        """

        indices = (np.argwhere(ctr_split >= 1.0)).flatten()
        years_of_max = self.project_years[indices]
        if len(years_of_max)>0:
            warnings.warn(
                f"The {fluid} contractor split equal more than 100% are in the following years {years_of_max} ",
                UserWarning)
            warnings.simplefilter("default", UserWarning)
        else:
            pass

        return years_of_max

    def run(self,
            sulfur_revenue: OtherRevenue = OtherRevenue.ADDITION_TO_GAS_REVENUE,
            electricity_revenue: OtherRevenue = OtherRevenue.ADDITION_TO_OIL_REVENUE,
            co2_revenue: OtherRevenue = OtherRevenue.ADDITION_TO_GAS_REVENUE,
            is_dmo_end_weighted=False,
            regime: GrossSplitRegime = GrossSplitRegime.PERMEN_ESDM_20_2019,
            tax_regime: TaxRegime = TaxRegime.NAILED_DOWN,
            tax_rate: float | np.ndarray = 0.22,
            sunk_cost_reference_year: int = None,
            depr_method: DeprMethod = DeprMethod.PSC_DB,
            decline_factor: float | int = 2,
            year_ref: int = None,
            tax_type: TaxType = TaxType.VAT,
            vat_rate: np.ndarray | float = 0.0,
            lbt_rate: np.ndarray | float = 0.0,
            inflation_rate: np.ndarray | float = 0.0,
            future_rate: float = 0.02,
            inflation_rate_applied_to: InflationAppliedTo | None = InflationAppliedTo.CAPEX,
            cum_production_split_offset: float | np.ndarray | None = 0.0,
            amortization: bool = False,
            sum_undepreciated_cost: bool = False
            ):

        # Configure Sunk Cost Reference Year
        if sunk_cost_reference_year is None:
            sunk_cost_reference_year = self.start_date.year

        if sunk_cost_reference_year > self.oil_onstream_date.year:
            raise SunkCostException(
                f"Sunk Cost Reference Year {sunk_cost_reference_year} "
                f"is after the on stream date: {self.oil_onstream_date}"
            )

        if sunk_cost_reference_year > self.gas_onstream_date.year:
            raise SunkCostException(
                f"Sunk Cost Reference Year {sunk_cost_reference_year} "
                f"is after the on stream date: {self.gas_onstream_date}"
            )

        if sunk_cost_reference_year < self.start_date.year:
            raise SunkCostException(
                f"Sunk Cost Reference Year {sunk_cost_reference_year} "
                f"is before the project start date: {self.start_date}"
            )

        if sunk_cost_reference_year > self.end_date.year:
            raise SunkCostException(
                f"Sunk Cost Reference Year {sunk_cost_reference_year} "
                f"is after the project end date: {self.end_date}"
            )

        # Configure year reference
        if year_ref is None:
            year_ref = self.start_date.year

        # Checking if the Cumulative Production Split Offset length is same with the project years
        if isinstance(cum_production_split_offset, np.ndarray):
            if len(cum_production_split_offset) != len(self.project_years):
                raise CumulativeProductionSplitException(
                    f"Length of the cum_production_split_offset: {len(cum_production_split_offset)} "
                    f"is not the same with Length of the project years: {len(self.project_years)}"
                )

        # Get the WAP Price
        self._get_wap_price()

        # Calculate expenditures for every cost components
        self._get_expenditures(
            year_ref=year_ref,
            tax_type=tax_type,
            vat_rate=vat_rate,
            lbt_rate=lbt_rate,
            inflation_rate=inflation_rate,
            future_rate=future_rate,
            inflation_rate_applied_to=inflation_rate_applied_to
        )

        # Get The Other Revenue as the chosen selection
        self._get_other_revenue(sulfur_revenue=sulfur_revenue,
                                electricity_revenue=electricity_revenue,
                                co2_revenue=co2_revenue)

        # Depreciation (tangible cost)
        (
            self._oil_depreciation,
            self._oil_undepreciated_asset,
        ) = self._oil_capital_cost.total_depreciation_rate(
            depr_method=depr_method,
            decline_factor=decline_factor,
            year_ref=year_ref,
            tax_type=tax_type,
            vat_rate=vat_rate,
            lbt_rate=lbt_rate,
            inflation_rate=inflation_rate,
        )

        (
            self._gas_depreciation,
            self._gas_undepreciated_asset,
        ) = self._gas_capital_cost.total_depreciation_rate(
            depr_method=depr_method,
            decline_factor=decline_factor,
            year_ref=year_ref,
            tax_type=tax_type,
            vat_rate=vat_rate,
            lbt_rate=lbt_rate,
            inflation_rate=inflation_rate,
        )

        # Treatment for small order of number, in example 1e-15
        self._oil_undepreciated_asset = np.where(
            self._oil_undepreciated_asset < 1.0e-5, 0, self._oil_undepreciated_asset)
        self._gas_undepreciated_asset = np.where(
            self._gas_undepreciated_asset < 1.0e-5, 0, self._gas_undepreciated_asset)

        # Treatment of the un-depreciated asset to be summed up in the last year of the contract or not
        if sum_undepreciated_cost is True:
            self._oil_depreciation[-1] = self._oil_depreciation[-1] + self._oil_undepreciated_asset
            self._gas_depreciation[-1] = self._gas_depreciation[-1] + self._gas_undepreciated_asset
        else:
            pass

        # Non Capital Cost
        self._oil_non_capital = (
                self._oil_intangible_expenditures
                + self._oil_opex_expenditures
                + self._oil_asr_expenditures
        )

        self._gas_non_capital = (
                self._gas_intangible_expenditures
                + self._gas_opex_expenditures
                + self._gas_asr_expenditures
        )

        # Get Sunk Cost
        self._get_sunk_cost(sunk_cost_reference_year)

        # Amortization Cost
        if amortization is True:
            self._oil_depreciation += unit_of_production_rate(
                start_year_project=self.start_date.year,
                cost=float(np.sum(self._oil_sunk_cost)),
                prod=self._oil_lifting.get_lifting_rate_arr(),
                prod_year=self.project_years,
                salvage_value=0.0,
                amortization_len=self.project_duration,)

            self._gas_depreciation += unit_of_production_rate(
                start_year_project=self.start_date.year,
                cost=float(np.sum(self._gas_sunk_cost)),
                prod=self._gas_lifting.get_lifting_rate_arr(),
                prod_year=self.project_years,
                salvage_value=0.0,
                amortization_len=self.project_duration,)

        else:
            pass

        # Variable Split. -> Will set the value of _variable_split
        self._wrapper_variable_split(regime=regime)

        # Base Split
        self._oil_base_split = np.full_like(self.project_years, fill_value=self.base_split_ctr_oil, dtype=float)
        self._gas_base_split = np.full_like(self.project_years, fill_value=self.base_split_ctr_gas, dtype=float)

        # Variable Split
        self._var_split_array = np.full_like(self.project_years, fill_value=self._variable_split, dtype=float)

        # Calculating the gas production in MMBOE
        # Based on the discussion on 28 of June 2024, with YRA in Whatsapp,
        # The cum. production used for progressive cum. production split is the sum of production rate and baseline

        prod_gas_boe = self._gas_lifting.get_prod_rate_total_arr() / self.conversion_boe_to_scf

        # Condition when the offset cumulative production array when user input is float
        if isinstance(cum_production_split_offset, float) or isinstance(cum_production_split_offset, int):
            offset_arr = np.full_like(self.project_years, fill_value=0.0, dtype=float)
            offset_arr[0] = cum_production_split_offset

        else:
            offset_arr = np.full_like(self.project_years, fill_value=0.0, dtype=float)

        # Calculating the cumulative production
        self._cumulative_prod = np.cumsum(self._oil_lifting.get_prod_rate_total_arr() + prod_gas_boe + offset_arr)

        # Progressive Split
        vectorized_get_prog_split = np.vectorize(self._wrapper_progressive_split)

        # Condition when the cum_production_split_offset is filled with np.ndarray
        if isinstance(cum_production_split_offset, np.ndarray) and len(cum_production_split_offset) > 1:
            oil_prog_price_split, oil_prog_cum_split = vectorized_get_prog_split(
                fluid=self._oil_lifting.fluid_type,
                price=self._oil_lifting.get_price_arr(),
                cum=None,
                regime=regime
            )

            self._oil_prog_split = oil_prog_price_split + cum_production_split_offset

            gas_prog_price_split, gas_prog_cum_split = vectorized_get_prog_split(
                fluid=self._gas_lifting.fluid_type,
                price=self._gas_lifting.get_price_arr(),
                cum=None,
                regime=regime
            )

            self._gas_prog_split = gas_prog_price_split + cum_production_split_offset

        # Condition when the cum_production_split_offset is not filled
        else:
            oil_prog_price_split, oil_prog_cum_split = vectorized_get_prog_split(
                fluid=self._oil_lifting.fluid_type,
                price=self._oil_lifting.get_price_arr(),
                cum=self._cumulative_prod,
                regime=regime
            )
            self._oil_prog_split = oil_prog_price_split + oil_prog_cum_split

            gas_prog_price_split, gas_prog_cum_split = vectorized_get_prog_split(
                fluid=self._gas_lifting.fluid_type,
                price=self._gas_lifting.get_price_arr(),
                cum=self._cumulative_prod,
                regime=regime
            )
            self._gas_prog_split = gas_prog_price_split + gas_prog_cum_split

        # Ministerial Discretion
        minis_disc_array = np.full_like(self.project_years, fill_value=self.split_ministry_disc, dtype=float)

        # Total Contractor Split
        self._oil_ctr_split = (self._oil_base_split + self._var_split_array + self._oil_prog_split +
                               minis_disc_array)
        self._gas_ctr_split = (self._gas_base_split + self._var_split_array + self._gas_prog_split +
                               minis_disc_array)

        self._oil_ctr_split_prior_bracket = np.copy(self._oil_ctr_split)
        self._gas_ctr_split_prior_bracket = np.copy(self._gas_ctr_split)

        # Add the condition to show the contractor split more than 100%
        self._oil_year_maximum_ctr_split = self._get_year_maximum_split(
            ctr_split=self._oil_ctr_split,
            fluid="Oil"
        )

        self._gas_year_maximum_ctr_split = self._get_year_maximum_split(
            ctr_split=self._gas_ctr_split,
            fluid="Gas"
        )

        # Condition to limit the contractor split is 1.0 at maximum
        self._oil_ctr_split = np.where(
            self._oil_ctr_split > np.full_like(self._oil_ctr_split, 1.0),
            np.full_like(self._oil_ctr_split, 1.0),
            self._oil_ctr_split
        )

        self._gas_ctr_split = np.where(
            self._gas_ctr_split > np.full_like(self._gas_ctr_split, 1.0),
            np.full_like(self._gas_ctr_split, 1.0),
            self._gas_ctr_split
        )

        # Contractor Share
        self._oil_ctr_share_before_transfer = self._oil_revenue * self._oil_ctr_split
        self._gas_ctr_share_before_transfer = self._gas_revenue * self._gas_ctr_split

        # Government Share
        self._oil_gov_share = self._oil_revenue - self._oil_ctr_share_before_transfer
        self._gas_gov_share = self._gas_revenue - self._gas_ctr_share_before_transfer

        # Total Investment
        self._oil_total_expenses = (self._oil_capital_expenditures + self._oil_intangible_expenditures +
                                    self._oil_opex_expenditures + self._oil_asr_expenditures)
        self._gas_total_expenses = (self._gas_capital_expenditures + self._gas_intangible_expenditures +
                                    self._gas_opex_expenditures + self._gas_asr_expenditures)

        # Cost to be Deducted
        self._oil_cost_tobe_deducted = (self._oil_depreciation + self._oil_intangible_expenditures +
                                        self._oil_opex_expenditures + self._oil_asr_expenditures)
        self._gas_cost_tobe_deducted = (self._gas_depreciation + self._gas_intangible_expenditures +
                                        self._gas_opex_expenditures + self._gas_asr_expenditures)

        # Carry Forward Deductible Cost (In PSC Cost Recovery called Unrecovered Cost)
        self._oil_carward_deduct_cost = psc_tools.get_unrecovered_cost(depreciation=self._oil_depreciation,
                                                                       non_capital=self._oil_non_capital,
                                                                       revenue=self._oil_ctr_share_before_transfer,
                                                                       ftp_ctr=np.zeros_like(self.project_years),
                                                                       ftp_gov=np.zeros_like(self.project_years),
                                                                       ic=np.zeros_like(self.project_years))

        self._gas_carward_deduct_cost = psc_tools.get_unrecovered_cost(depreciation=self._gas_depreciation,
                                                                       non_capital=self._gas_non_capital,
                                                                       revenue=self._gas_ctr_share_before_transfer,
                                                                       ftp_ctr=np.zeros_like(self.project_years),
                                                                       ftp_gov=np.zeros_like(self.project_years),
                                                                       ic=np.zeros_like(self.project_years))

        # Deductible Cost (In PSC Cost Recovery called Cost Recovery)
        self._oil_deductible_cost = self._get_deductible_cost(ctr_gross_share=self._oil_ctr_share_before_transfer,
                                                              cost_tobe_deducted=self._oil_cost_tobe_deducted,
                                                              carward_deduct_cost=self._oil_carward_deduct_cost)

        self._gas_deductible_cost = self._get_deductible_cost(ctr_gross_share=self._gas_ctr_share_before_transfer,
                                                              cost_tobe_deducted=self._gas_cost_tobe_deducted,
                                                              carward_deduct_cost=self._gas_carward_deduct_cost)

        # Transfer
        self._transfer_to_oil, self._transfer_to_gas = psc_tools.get_transfer(
            oil_unrecovered=self._oil_carward_deduct_cost,
            gas_unrecovered=self._gas_carward_deduct_cost,
            oil_ets_pretransfer=self._oil_ctr_share_before_transfer,
            gas_ets_pretransfer=self._gas_ctr_share_before_transfer)

        # Carry Forward Deductible Cost After Transfer
        self._oil_carward_cost_aftertf = self._oil_carward_deduct_cost - self._transfer_to_gas
        self._gas_carward_cost_aftertf = self._gas_carward_deduct_cost - self._transfer_to_oil

        # Contractor Share After Transfer
        self._oil_ctr_share_after_transfer = (self._oil_ctr_share_before_transfer +
                                              self._transfer_to_oil -
                                              self._transfer_to_gas)

        self._gas_ctr_share_after_transfer = (self._gas_ctr_share_before_transfer +
                                              self._transfer_to_gas -
                                              self._transfer_to_oil)

        # Contractor Net Operating Profit
        self._oil_net_operating_profit = self._oil_ctr_share_after_transfer - self._oil_deductible_cost
        self._gas_net_operating_profit = self._gas_ctr_share_after_transfer - self._gas_deductible_cost

        # DMO
        self._oil_dmo_volume, self._oil_dmo_fee, self._oil_ddmo = psc_tools.get_dmo_gross_split(
            onstream_date=self.oil_onstream_date,
            start_date=self.start_date,
            project_years=self.project_years,
            dmo_holiday_duration=self.oil_dmo_holiday_duration,
            dmo_volume_portion=self.oil_dmo_volume_portion,
            dmo_fee_portion=self.oil_dmo_fee_portion,
            price=self._oil_wap_price,
            unrecovered_cost=self._oil_carward_cost_aftertf,
            is_dmo_end_weighted=is_dmo_end_weighted,
            net_operating_profit=self._oil_net_operating_profit,
            contractor_share=self._oil_ctr_share_after_transfer,)

        self._gas_dmo_volume, self._gas_dmo_fee, self._gas_ddmo = psc_tools.get_dmo_gross_split(
            onstream_date=self.gas_onstream_date,
            start_date=self.start_date,
            project_years=self.project_years,
            dmo_holiday_duration=self.gas_dmo_holiday_duration,
            dmo_volume_portion=self.gas_dmo_volume_portion,
            dmo_fee_portion=self.gas_dmo_fee_portion,
            price=self._gas_wap_price,
            unrecovered_cost=self._gas_carward_cost_aftertf,
            is_dmo_end_weighted=is_dmo_end_weighted,
            net_operating_profit=self._gas_net_operating_profit,
            contractor_share=self._gas_ctr_share_after_transfer,)

        # Taxable Income
        self._oil_taxable_income = self._oil_net_operating_profit - self._oil_ddmo
        self._gas_taxable_income = self._gas_net_operating_profit - self._gas_ddmo

        # Tax Payment
        # Generating Tax array if tax_rate argument is a single value not array
        if isinstance(tax_rate, float) or isinstance(tax_rate, int):
            self._tax_rate_arr = np.full_like(self.project_years, tax_rate, dtype=float)

        # Generating Tax array based on the tax regime if tax_rate argument is None
        if tax_rate is None:
            self._tax_rate_arr = self._get_tax_by_regime(tax_regime=tax_regime)

        elif isinstance(tax_rate, np.ndarray):
            self._tax_rate_arr = tax_rate

        self._oil_tax = self._oil_taxable_income * self._tax_rate_arr
        self._gas_tax = self._gas_taxable_income * self._tax_rate_arr

        # Contractor Net Share
        self._oil_ctr_net_share = self._oil_taxable_income - self._oil_tax
        self._gas_ctr_net_share = self._gas_taxable_income - self._gas_tax

        # Contractor Cash Flow
        self._oil_ctr_cashflow = (self._oil_ctr_share_before_transfer - self._oil_total_expenses - self._oil_ddmo -
                                  self._oil_tax)
        self._gas_ctr_cashflow = (self._gas_ctr_share_before_transfer - self._gas_total_expenses - self._gas_ddmo -
                                  self._gas_tax)

        # Government Take
        self._oil_government_take = self._oil_gov_share + self._oil_ddmo + self._oil_tax
        self._gas_government_take = self._gas_gov_share + self._gas_ddmo + self._gas_tax

        # Consolidated attributes
        self._consolidated_revenue = self._oil_revenue + self._gas_revenue
        self._consolidated_capital_cost = self._oil_capital_expenditures + self._gas_capital_expenditures
        self._consolidated_intangible = self._oil_intangible_expenditures + self._gas_intangible_expenditures
        self._consolidated_sunk_cost = self._oil_sunk_cost + self._gas_sunk_cost
        self._consolidated_opex = self._oil_opex_expenditures + self._gas_opex_expenditures
        self._consolidated_asr = self._oil_asr_expenditures + self._gas_asr_expenditures
        self._consolidated_non_capital = self._oil_non_capital + self._gas_non_capital
        self._consolidated_depreciation = self._oil_depreciation + self._gas_depreciation
        self._consolidated_undepreciated_asset = self._oil_undepreciated_asset + self._gas_undepreciated_asset
        self._consolidated_ctr_share_before_tf = self._oil_ctr_share_before_transfer + self._gas_ctr_share_before_transfer
        self._consolidated_gov_share_before_tf = self._oil_gov_share + self._gas_gov_share
        self._consolidated_total_expenses = self._oil_total_expenses + self._gas_total_expenses
        self._consolidated_cost_tobe_deducted = self._oil_cost_tobe_deducted + self._gas_cost_tobe_deducted
        self._consolidated_carward_deduct_cost = self._oil_carward_deduct_cost + self._gas_carward_deduct_cost
        self._consolidated_deductible_cost = self._oil_deductible_cost + self._gas_deductible_cost
        self._consolidated_carward_cost_aftertf = self._oil_carward_cost_aftertf + self._gas_carward_cost_aftertf
        self._consolidated_ctr_share_after_transfer = self._oil_ctr_share_after_transfer + self._gas_ctr_share_after_transfer
        self._consolidated_net_operating_profit = self._oil_net_operating_profit + self._gas_net_operating_profit
        self._consolidated_dmo_volume = self._oil_dmo_volume + self._gas_dmo_volume
        self._consolidated_dmo_fee = self._oil_dmo_fee + self._gas_dmo_fee
        self._consolidated_ddmo = self._oil_ddmo + self._gas_ddmo
        self._consolidated_taxable_income = self._oil_taxable_income + self._gas_taxable_income
        self._consolidated_tax_payment = self._oil_tax + self._gas_tax
        self._consolidated_ctr_net_share = self._oil_ctr_net_share + self._gas_ctr_net_share
        self._consolidated_government_take = self._oil_government_take + self._gas_government_take
        self._consolidated_cashflow = self._oil_ctr_cashflow + self._gas_ctr_cashflow

    def __len__(self):
        return self.project_duration

    def __eq__(self, other):
        # Between two instances of Gross Split
        if isinstance(other, GrossSplit):
            return all(
                (
                    np.allclose(self._oil_lifting.lifting_rate, other._oil_lifting.lifting_rate),
                    np.allclose(self._gas_lifting.lifting_rate, other._gas_lifting.lifting_rate),
                    np.allclose(self._oil_revenue, other._oil_revenue),
                    np.allclose(self._gas_revenue, other._gas_revenue),
                    np.allclose(self._oil_capital_expenditures, other._oil_capital_expenditures),
                    np.allclose(self._gas_capital_expenditures, other._gas_capital_expenditures),
                    np.allclose(self._oil_intangible_expenditures, other._oil_intangible_expenditures),
                    np.allclose(self._gas_intangible_expenditures, other._gas_intangible_expenditures),
                    np.allclose(self._oil_opex_expenditures, other._oil_opex_expenditures),
                    np.allclose(self._gas_opex_expenditures, other._gas_opex_expenditures),
                    np.allclose(self._oil_asr_expenditures, other._oil_asr_expenditures),
                    np.allclose(self._gas_asr_expenditures, other._gas_asr_expenditures),
                )
            )

        else:
            return False

    def __lt__(self, other):
        if isinstance(other, GrossSplit):

            self.run()
            other.run()

            self_total_base_cashflow = self._oil_base_cashflow + self._gas_base_cashflow
            other_total_base_cashflow = other._oil_base_cashflow + other._gas_base_cashflow

            return self_total_base_cashflow.irr() < other_total_base_cashflow.irr()

        else:
            raise GrossSplitException

    def __le__(self, other):
        if isinstance(other, GrossSplit):
            self.run()
            other.run()

            self_total_base_cashflow = self._oil_base_cashflow + self._gas_base_cashflow
            other_total_base_cashflow = other._oil_base_cashflow + other._gas_base_cashflow

            return self_total_base_cashflow.irr() <= other_total_base_cashflow.irr()

        else:
            raise GrossSplitException

    def __gt__(self, other):
        if isinstance(other, GrossSplit):

            self.run()
            other.run()

            self_total_base_cashflow = self._oil_base_cashflow + self._gas_base_cashflow
            other_total_base_cashflow = other._oil_base_cashflow + other._gas_base_cashflow

            return self_total_base_cashflow.irr() > other_total_base_cashflow.irr()

        else:
            raise GrossSplitException

    def __ge__(self, other):
        if isinstance(other, GrossSplit):

            self.run()
            other.run()

            self_total_base_cashflow = self._oil_base_cashflow + self._gas_base_cashflow
            other_total_base_cashflow = other._oil_base_cashflow + other._gas_base_cashflow

            return self_total_base_cashflow.irr() >= other_total_base_cashflow.irr()

        else:
            raise GrossSplitException



""" Specify callable classes and methods from package 'contracts' """

from .project import BaseProject
from .costrecovery import CostRecovery
from .grossplit import GrossSplit
from .transition import Transition
